"""Defines FieldOfView class"""
from copy import deepcopy

import numpy as np
from scipy.interpolate import interp1d

from astropy import units as u
from astropy.io import fits
from astropy.table import Table
from synphot import Empirical1D, SourceSpectrum
from synphot.units import PHOTLAM

from . import fov_utils as fu
from . import image_plane_utils as imp_utils

from ..base_classes import SourceBase, FieldOfViewBase
from .. import utils


class FieldOfView(FieldOfViewBase):
    """
    A FOV is spectro-spatial volume cut out of a Source object.

    Flux units after extracting the fields from the Source are in ph/s/pixel

    The initial header should contain an on-sky WCS description:
    - CDELT-, CUNIT-, NAXIS- : for pixel scale and size (assumed CUNIT in deg)
    - CRVAL-, CRPIX- : for positioning the final image
    - CTYPE- : is assumed to be "RA---TAN", "DEC---TAN"

    and an image-plane WCS description
    - CDELT-D, CUNIT-D, NAXISn : for pixel scale and size (assumed CUNIT in mm)
    - CRVAL-D, CRPIX-D : for positioning the final image
    - CTYPE-D : is assumed to be "LINEAR", "LINEAR"

    The wavelength range is given by waverange

    """

    def __init__(self, header, waverange, detector_header=None, **kwargs):
        self.meta = {"id": None,
                     "wave_min": utils.quantify(waverange[0], u.um),
                     "wave_max": utils.quantify(waverange[1], u.um),
                     "wave_bin_n": 1,
                     "wave_bin_type": "linear",

                     "area": 0 * u.m**2,
                     "pixel_area": None,                    # [arcsec]
                     "sub_pixel": "!SIM.sub_pixel.flag",
                     "distortion": {"scale": [1, 1],
                                    "offset": [0, 0],
                                    "shear": [1, 1],
                                    "rotation": 0,
                                    "radius_of_curvature": None},
                     "conserve_image": True,
                     "trace_id": None,
                     "aperture_id": None,
                     }
        self.meta.update(kwargs)

        if not any((utils.has_needed_keywords(header, s) for s in ["", "S"])):
            raise ValueError(f"header must contain a valid sky-plane WCS: {dict(header)}")
        if not utils.has_needed_keywords(header, "D"):
            raise ValueError(f"header must contain a valid image-plane WCS: {dict(header)}")

        self.header = fits.Header()
        self.header["NAXIS"] = 2
        self.header["NAXIS1"] = header["NAXIS1"]
        self.header["NAXIS2"] = header["NAXIS2"]
        self.header.update(header)
        self.detector_header = detector_header
        self.hdu = None

        self.image_plane_id = 0
        self.fields = []
        self.spectra = {}

        self.cube = None        # 3D array for IFU, long-lit, Slicer-MOS
        self.image = None       # 2D array for Imagers
        self.spectrum = None    # SourceSpectrum for Fibre-fed MOS

        self._waverange = None
        self._wavelength = None
        self._volume = None

    @property
    def pixel_area(self):
        if self.meta["pixel_area"] is None:
            hdr = self.header
            pixarea = (hdr['CDELT1'] * u.Unit(hdr['CUNIT1']) *
                       hdr['CDELT2'] * u.Unit(hdr['CUNIT2'])).to(u.arcsec ** 2)
            self.meta["pixel_area"] = pixarea.value     # [arcsec]

        return self.meta["pixel_area"]

    def extract_from(self, src):
        """ ..assumption: Bandpass has been applied

        .. note:: Spectra are cut and copied from the original Source object.
            They are in original units. ph/s/pix comes in the make_**** methods

        """

        if not isinstance(src, SourceBase):
            raise ValueError(f"source must be a Source object: {type(src)}")

        fields_in_fov = [field for field in src.fields
                         if fu.is_field_in_fov(self.header, field)]

        spec_refs = []
        volume = self.volume()
        for ifld, fld in enumerate(fields_in_fov):
            if isinstance(fld, Table):
                fields_in_fov[ifld] = fu.extract_area_from_table(fld, volume)
                spec_refs += list(np.unique(fields_in_fov[ifld] ["ref"]))

            elif isinstance(fld, fits.ImageHDU):
                if fld.header["NAXIS"] in (2, 3):
                    fields_in_fov[ifld] = fu.extract_area_from_imagehdu(fld, volume)
                if fld.header["NAXIS"] == 2 or fld.header.get("BG_SRC"):
                    ref = fld.header.get("SPEC_REF")
                    if ref is not None:
                        spec_refs += [ref]

        waves = volume["waves"] * u.Unit(volume["wave_unit"])
        spectra = {ref: fu.extract_range_from_spectrum(src.spectra[ref], waves)
                   for ref in np.unique(spec_refs)}

        self.fields = fields_in_fov
        self.spectra = spectra

    def view(self, hdu_type="image", sub_pixel=None, use_photlam=None):
        """
        Forces the self.fields to be viewed as a single object

        Parameters
        ----------
        sub_pixel : bool
        hdu_type : str
            ["cube", "image", "spectrum"]

        Returns
        -------
        self.hdu : fits.ImageHDU, synphot.SourceSpectrum

        """
        if sub_pixel is not None:
            self.meta["sub_pixel"] = sub_pixel

        if hdu_type == "image":
            use_photlam = False if use_photlam is None else use_photlam
            self.hdu = self.make_image_hdu(use_photlam=use_photlam)
        elif hdu_type == "cube":
            self.hdu = self.make_cube_hdu()
        elif hdu_type == "spectrum":
            self.hdu = self.make_spectrum()

        return self.hdu

    def flatten(self):
        if self.hdu and self.hdu.header["NAXIS"] == 3:
            image = np.sum(self.hdu.data, axis=0)
            self.hdu.data = image

    def make_spectrum(self):
        """
        This is needed for when we do incoherent MOS instruments.
        Each fibre doesn't care about the spatial information.

        1. Make waveset and zeros flux
            make dict of spectra evaluated at waveset

        2. Find Cube fields
            collapse cube along spatial dimensions --> spectrum vector
            convert vector to PHOTLAM
            interpolate at waveset
            add to zeros flux vector

        3. Find Image fields
            sum image over both dimensions
            evaluate SPEC_REF spectum at waveset
            multiply by sum
            add to zeros flux vector

        4. Find Table fields
            evaluate all spectra at waveset
            for each unique ref, sum the weights
            add each spectra * sum of weights to zeros flux vector

        Returns
        -------
        spec : SourceSpectrum
            [PHOTLAM]

        """
        fov_waveset = self.waveset
        canvas_flux = np.zeros(len(fov_waveset))

        specs = {ref: spec(fov_waveset) for ref, spec in self.spectra.items()}

        for field in self.cube_fields:
            hdu_waveset = fu.get_cube_waveset(field.header, return_quantity=True)
            fluxes = field.data.sum(axis=2).sum(axis=1)
            fov_waveset_fluxes = np.interp(fov_waveset, hdu_waveset, fluxes)

            field_unit = field.header.get("BUNIT", PHOTLAM)
            flux_scale_factor = u.Unit(field_unit).to(PHOTLAM)

            canvas_flux += fov_waveset_fluxes * flux_scale_factor

        for field in self.image_fields:
            ref = field.header["SPEC_REF"]
            weight = np.sum(field.data)
            canvas_flux += self.spectra[ref](fov_waveset).value * weight


        for field in self.table_fields:
            refs = np.array(field["ref"])
            weights = np.array(field["weight"])
            weight_sums = {ref: np.sum(weights[refs == ref])
                           for ref in np.unique(refs)}
            for ref, weight in weight_sums.items():
                canvas_flux += self.spectra[ref](fov_waveset).value * weight

        for field in self.background_fields:
            bg_solid_angle = u.Unit(field.header["SOLIDANG"]).to(u.arcsec**-2)
            area_factor = self.pixel_area * bg_solid_angle       # arcsec**2 * arcsec**-2

            ref = field.header["SPEC_REF"]
            canvas_flux += self.spectra[ref](fov_waveset).value * area_factor

        spectrum = SourceSpectrum(Empirical1D, points=fov_waveset,
                                  lookup_table=canvas_flux)

        return spectrum

    def make_image_hdu(self, use_photlam=False):
        """
        Used for imaging

        Output image units are ph s-1 pixel-1

        .. note:: ``self.make_image()`` does NOT store anything in ``self.image``

            See make_cube for an explanation

        1. Make waveset and canvas image
            make canvas image from NAXIS1,2 from fov.header

        2. Find Cube fields
            collapse cube along wavelength axis
            rescale and reorient image
            add cube image to canvas image

        3. Find Image fields
            rescale and reorient images
            sum spectra between wavelength edges
            multiply image by summed spectrum
            add image to canvas image

        4. Find Table fields
            sum spectra between wavelength edges
            add summed flux at x,y position in canvas image

        Parameters
        ----------
        use_photlam : bool
            Default False. Defines the flux units of the image pixels

        Returns
        -------
        image_hdu : fits.ImageHDU
            [ph s-1 pixel-1] or PHOTLAM (if use_photlam=True)

        """
        spline_order = utils.from_currsys("!SIM.computing.spline_order")

        # 1. Make waveset and canvas image
        fov_waveset = np.unique(self.waveset)
        bin_widths = np.diff(fov_waveset)       # u.um
        bin_widths = 0.5 * (np.r_[0, bin_widths] + np.r_[bin_widths, 0])
        area = utils.from_currsys(self.meta["area"])    # u.m2

        # PHOTLAM * u.um * u.m2 --> ph / s
        specs = {ref: spec(fov_waveset) for ref, spec in self.spectra.items()}
        if use_photlam is False:
            for key in specs:
                specs[key] = (specs[key] * bin_widths * area).to(u.ph / u.s)

        fluxes = {ref: np.sum(spec).value for ref, spec in specs.items()}
        naxis1, naxis2 = self.header["NAXIS1"], self.header["NAXIS2"]
        canvas_image_hdu = fits.ImageHDU(data=np.zeros((naxis2, naxis1)),
                                         header=self.header)

        # 2. Find Cube fields
        for field in self.cube_fields:
            # cube_fields come in with units of photlam/arcsec2, need to convert to ph/s
            # We need to the voxel volume (spectral and solid angle) for that.
            # ..todo: implement branch for use_photlam is True
            spectral_bin_width = (field.header['CDELT3'] *
                                  u.Unit(field.header['CUNIT3'])).to(u.Angstrom)
            pixarea = (field.header['CDELT1'] * u.Unit(field.header['CUNIT1']) *
                       field.header['CDELT2'] * u.Unit(field.header['CUNIT2'])).to(u.arcsec**2)

            # First collapse to image, then convert units
            image = np.sum(field.data, axis=0) * PHOTLAM/u.arcsec**2
            image = (image * pixarea * area * spectral_bin_width).to(u.ph/u.s)

            tmp_hdu = fits.ImageHDU(data=image, header=field.header)
            canvas_image_hdu = imp_utils.add_imagehdu_to_imagehdu(
                tmp_hdu,
                canvas_image_hdu,
                conserve_flux=True,
                spline_order=spline_order)

        # 2. Find Image fields
        for field in self.image_fields:
            image = deepcopy(field.data)
            tmp_hdu = fits.ImageHDU(data=image, header=field.header)
            tmp_hdu.data *= fluxes[field.header["SPEC_REF"]]  # ph / s
            canvas_image_hdu = imp_utils.add_imagehdu_to_imagehdu(
                tmp_hdu,
                canvas_image_hdu,
                conserve_flux=True,
                spline_order=spline_order)

        # 3. Find Table fields
        for field in self.table_fields:
            # x, y are ALWAYS in arcsec - crval is in deg
            xpix, ypix = imp_utils.val2pix(self.header,
                                           field["x"] / 3600,
                                           field["y"] / 3600)
            if utils.from_currsys(self.meta["sub_pixel"]):
                for i, row in enumerate(field):
                    xs, ys, fracs = imp_utils.sub_pixel_fractions(xpix[i],
                                                                  ypix[i])
                    ref, weight = row["ref"], row["weight"]
                    for x, y, f in zip(xs, ys, fracs):
                        canvas_image_hdu.data[y, x] += fluxes[ref] * weight * f
            else:
                # Note: these had x/ypix+0.5 until a06ab75
                x = np.array(xpix).astype(int)
                y = np.array(ypix).astype(int)     # quickest way to round
                f = np.array([fluxes[ref] for ref in field["ref"]])
                weight = np.array(field["weight"])

                # Mask out any stars that were pushed out of the fov by rounding
                m = (x < canvas_image_hdu.data.shape[1]) * \
                    (y < canvas_image_hdu.data.shape[0])
                canvas_image_hdu.data[y[m], x[m]] += f[m] * weight[m]

        # 4. Find Background fields
        for field in self.background_fields:
            bg_solid_angle = u.Unit(field.header["SOLIDANG"]).to(u.arcsec**-2)
            area_factor = self.pixel_area * bg_solid_angle       # arcsec**2 * arcsec**-2

            flux = fluxes[field.header["SPEC_REF"]] * area_factor
            canvas_image_hdu.data += flux

        image_hdu = canvas_image_hdu        # [ph s-1 pixel-1]

        return image_hdu

    def make_cube_hdu(self):
        """
        Used for IFUs, slit spectrographs, and coherent MOSs (e.g.KMOS)

        Returned cube units are ``ph s-1 voxel-1``

        .. note:: ``self.make_cube()`` does NOT store anything in ``self.cube``

            self.cube and self.make_cube() are deliberately kept seperately
            so that self.cube will not be accidently overwritten by a rogue
            call from an Effect object.

            All Effect objects should specifically test whether
            ``self.cube is None`` before assigning a new cube it

        The cube is made with these steps:

        1. Make waveset and canvas cube::

            if at least one cube:
                set waveset to equal largest cube waveset
            else:
                make waveset from self.meta values
            make canvas cube based on waveset of largest cube and NAXIS1,2 from fov.header

        2. Find Cube fields::

            rescale and reorient cubes
            interp1d smaller cubes with waveset
            add cubes to cavas cube

        3. Find Image fields::

            rescale and reorient images
            evaluate spectra at waveset
            expand image by spectra to 3D form
            add image cubes to canvas cube

        4. Find Table fields::

            evaluate spectra at waveset
            add spectrum at x,y position in canvas cube

        ``PHOTLAM = ph/s/m2/um``.
        Original source fields are in units of:

        - tables: (PHOTLAM in spectrum)
        - images: arcsec-2 (PHOTLAM in spectrum)
        - cubes: PHOTLAM arcsec-2

        .. warning:: Input Images and Cubes should have units of PHOTLAM arcsec-2

        Returns
        -------
        canvas_cube_hdu : fits.ImageHDU
            [ph s-1 AA-1 arcsec-2]      # as needed by SpectralTrace

        """
        spline_order = utils.from_currsys("!SIM.computing.spline_order")


        # 1. Make waveset and canvas cube (area, bin_width are applied at end)

        wave_min = self.meta["wave_min"]        # Quantity [um]
        wave_max = self.meta["wave_max"]

        wave_unit = u.Unit(utils.from_currsys("!SIM.spectral.wave_unit"))
        dwave = utils.from_currsys("!SIM.spectral.spectral_bin_width")  # Not a quantity
        fov_waveset = np.arange(wave_min.value, wave_max.value, dwave) * wave_unit
        fov_waveset = fov_waveset.to(u.um)

        # fov_waveset = self.waveset
        # wave_bin_n = len(fov_waveset)
        # if "lin" in self.meta["wave_bin_type"]:
        #     fov_waveset = np.linspace(wave_min, wave_max, wave_bin_n)
        # elif "log" in self.meta["wave_bin_type"]:
        #     wmin, wmax = wave_min.to(u.um).value, wave_max.to(u.um).value
        #     fov_waveset = np.logspace(wmin, wmax, wave_bin_n)

        specs = {ref: spec(fov_waveset)                     # PHOTLAM = ph/s/cm2/AA
                 for ref, spec in self.spectra.items()}

        # make canvas cube based on waveset of largest cube and NAXIS1,2 from fov.header
        naxis1, naxis2 = self.header["NAXIS1"], self.header["NAXIS2"]
        naxis3 = len(fov_waveset)
        canvas_cube_hdu = fits.ImageHDU(data=np.zeros((naxis3, naxis2, naxis1)),
                                        header=self.header)
        canvas_cube_hdu.header["BUNIT"] = "ph s-1 cm-2 AA-1"

        # 2. Add Cube fields
        for field in self.cube_fields:
            # Cube should be in PHOTLAM arcsec-2 for SpectralTrace mapping
            # Assumption is that ImageHDUs have units of PHOTLAM arcsec-2
            field_waveset = fu.get_cube_waveset(field.header,
                                                return_quantity=True)

            # ..todo: Deal with this bounds_error in a more elegant way
            field_interp = interp1d(field_waveset.to(u.um).value,
                                    field.data, axis=0, kind="linear",
                                    bounds_error=False, fill_value=0)

            field_data = field_interp(fov_waveset.value)

            # Pixel scale conversion
            field_pixarea = (field.header['CDELT1']
                             * field.header['CDELT2']
                             * u.Unit(field.header['CUNIT1'])
                             * u.Unit(field.header['CUNIT2'])).to(u.arcsec**2)
            field_pixarea = field_pixarea.value
            field_data *= field_pixarea / self.pixel_area
            field_hdu = fits.ImageHDU(data=field_data, header=field.header)
            canvas_cube_hdu = imp_utils.add_imagehdu_to_imagehdu(
                field_hdu,
                canvas_cube_hdu,
                spline_order=spline_order)

        # 3. Find Image fields
        for field in self.image_fields:
            # Cube should be in PHOTLAM arcsec-2 for SpectralTrace mapping
            # Assumption is that ImageHDUs have units of PHOTLAM arcsec-2
            # ImageHDUs have photons/second/pixel.
            # ..todo: Add a catch to get ImageHDU with BUNITs
            canvas_image_hdu = fits.ImageHDU(data=np.zeros((naxis2, naxis1)),
                                             header=self.header)
            pixarea = (field.header['CDELT1'] * u.Unit(field.header['CUNIT1']) *
                       field.header['CDELT2'] * u.Unit(field.header['CUNIT2'])).to(u.arcsec**2)

            field.data = field.data / self.pixel_area
            canvas_image_hdu = imp_utils.add_imagehdu_to_imagehdu(field,
                                                    canvas_image_hdu,
                                                    spline_order=spline_order)
            spec = specs[field.header["SPEC_REF"]]
            field_cube = canvas_image_hdu.data[None, :, :] * spec[:, None, None]  # 2D * 1D -> 3D
            canvas_cube_hdu.data += field_cube.value

        # 4. Find Table fields
        for field in self.table_fields:
            # Cube should be in PHOTLAM arcsec-2 for SpectralTrace mapping
            # Point sources are in PHOTLAM per pixel
            # Point sources need to be scaled up by inverse pixel_area
            pixel_area = self.pixel_area
            for row in field:
                xsky, ysky = row["x"], row["y"]
                ref, weight = row["ref"], row["weight"]
                # x, y are ALWAYS in arcsec - crval is in deg
                xpix, ypix = imp_utils.val2pix(self.header, xsky / 3600, ysky / 3600)
                if utils.from_currsys(self.meta["sub_pixel"]):
                    xs, ys, fracs = imp_utils.sub_pixel_fractions(xpix, ypix)
                    for i, j, k in zip(xs, ys, fracs):
                        flux_vector = specs[ref].value * weight * k / pixel_area
                        canvas_cube_hdu.data[:, j, i] +=  flux_vector
                else:
                    x, y = int(xpix), int(ypix)
                    flux_vector = specs[ref].value * weight / pixel_area
                    canvas_cube_hdu.data[:, y, x] += flux_vector

        # 5. Add Background fields
        for field in self.background_fields:
            # bg_solid_angle = u.Unit(field.header["SOLIDANG"]).to(u.arcsec**-2)  # float [arcsec-2]
            # pixel_area = utils.from_currsys(self.meta["pixel_scale"]) ** 2      # float [arcsec2]
            # area_factor = pixel_area * bg_solid_angle                           # float [arcsec2 * arcsec-2]

            # Cube should be in PHOTLAM arcsec-2 for SpectralTrace mapping
            # spec = specs[field.header["SPEC_REF"]] * area_factor
            spec = specs[field.header["SPEC_REF"]]
            canvas_cube_hdu.data += spec[:, None, None].value

        # 6. Convert from PHOTLAM to ph/s/voxel
        #    PHOTLAM = ph/s/cm-2/AA
        #    area = m2, fov_waveset = um
        # SpectralTrace wants ph/s/um/arcsec2 --> get rid of m2, leave um
        area = utils.from_currsys(self.meta["area"])  # u.m2
        canvas_cube_hdu.data *= area.to(u.cm ** 2).value
        canvas_cube_hdu.data *= 1e4        # ph/s/AA/arcsec2 --> ph/s/um/arcsec2

        # bin_widths = np.diff(fov_waveset).to(u.AA).value
        # bin_widths = 0.5 * (np.r_[0, bin_widths] + np.r_[bin_widths, 0])
        # canvas_cube_hdu.data *= bin_widths[:, None, None]

        cdelt3 = np.diff(fov_waveset[:2])[0]
        canvas_cube_hdu.header.update({"CDELT3": cdelt3.to(u.um).value,
                                       "CRVAL3": fov_waveset[0].value,
                                       "CRPIX3": 1,
                                       "CUNIT3": "um",
                                       "CTYPE3": "WAVE"})
        # ..todo: Add the log wavelength keyword here, if log scale is needed

        cube_hdu = canvas_cube_hdu      # [ph s-1 AA-1 (arcsec-2)]

        return cube_hdu

    def volume(self, wcs_prefix=""):
        xs, ys = imp_utils.calc_footprint(self.header, wcs_suffix=wcs_prefix)
        wave_corners = self.waverange
        self._volume = {"xs": [min(xs), max(xs)],
                        "ys": [min(ys), max(ys)],
                        "waves": self.waverange,
                        "xy_unit": "mm" if wcs_prefix == "D" else "deg",
                        "wave_unit": "um"}
        return self._volume

    @property
    def data(self):
        if self.hdu is not None:
            data = self.hdu.data
        elif self.image is not None:
            data = self.image
        elif self.cube is not None:
            data = self.cube
        elif self.spectrum is not None:
            data = self.spectrum
        else:
            data = None

        return data

    @property
    def corners(self):
        sky_corners = imp_utils.calc_footprint(self.header)
        imp_corners = imp_utils.calc_footprint(self.header, "D")
        return sky_corners, imp_corners

    @property
    def waverange(self):
        """Returns wavelength range in um [wave_min, wave_max]"""
        if self._waverange is None:
            wave_min = utils.quantify(self.meta["wave_min"], u.um).value
            wave_max = utils.quantify(self.meta["wave_max"], u.um).value
            self._waverange = [wave_min, wave_max]
        return self._waverange

    @property
    def wavelength(self):
        """Returns central wavelength in um"""
        if self._wavelength is None:
            self._wavelength = np.average(self.waverange)
        return utils.quantify(self._wavelength, u.um)

    @property
    def waveset(self):
        """Returns a wavelength vector in um"""

        field_cubes = self.cube_fields
        if len(field_cubes) > 0:
            i = np.argmax([cube.header["NAXIS3"] for cube in field_cubes])
            _waveset = fu.get_cube_waveset(field_cubes[i].header,
                                           return_quantity=True)
        elif len(self.spectra) > 0:
            wavesets = [spec.waveset for spec in self.spectra.values()]
            _waveset = np.concatenate(wavesets)
        else:
            _waveset = self.waverange * u.um

        _waveset = np.unique(_waveset)

        return _waveset.to(u.um)

    @property
    def cube_fields(self):
        return [field for field in self.fields
                if isinstance(field, fits.ImageHDU)
                and field.header["NAXIS"] == 3
                and field.header.get("BG_SRC", False) is False]

    @property
    def image_fields(self):
        return [field for field in self.fields
                if isinstance(field, fits.ImageHDU)
                and field.header["NAXIS"] == 2
                and field.header.get("BG_SRC", False) is False]

    @property
    def table_fields(self):
        return [field for field in self.fields
                if isinstance(field, Table)]


    @property
    def background_fields(self):
        return [field for field in self.fields
                if isinstance(field, fits.ImageHDU)
                and field.header.get("BG_SRC", False) is True]

    def __repr__(self):
        msg = "FOV id: {}, with dimensions ({}, {})\n" \
              "".format(self.meta["id"], self.header["NAXIS1"],
                        self.header["NAXIS2"])
        msg += "Sky centre: ({}, {})\n" \
               "".format(self.header["CRVAL1"], self.header["CRVAL2"])
        msg += "Image centre: ({}, {})\n" \
               "".format(self.header["CRVAL1D"], self.header["CRVAL2D"])
        msg += "Wavelength range: ({}, {})um\n" \
               "".format(self.meta["wave_min"], self.meta["wave_max"])

        return msg



"""SpectralTraceList and SpectralTrace for the METIS LM spectrograph"""
from copy import deepcopy
import numpy as np
from scipy.interpolate import RectBivariateSpline

from astropy.io import fits
from astropy.io import ascii as ioascii
from astropy.table import Table
from astropy.wcs import WCS
from astropy import units as u

from ..utils import from_currsys, find_file, quantify
from .spectral_trace_list import SpectralTraceList
from .spectral_trace_list_utils import SpectralTrace
from .spectral_trace_list_utils import Transform2D
from .apertures import ApertureMask
from .ter_curves import TERCurve
from ..base_classes import FieldOfViewBase, FOVSetupBase
from ..optics.fov import FieldOfView


class MetisLMSSpectralTraceList(SpectralTraceList):
    """
    SpectralTraceList for the METIS LM spectrograph
    """
    _class_params = {
        "naxis1": 122,
        "nslice": 28,
        "slicewidth": 0.0207,  # arcsec
        "pixscale": 0.0082,    # arcsec
        "grat_spacing": 18.2,
        "plate_scale": 0.303,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        #self.params = {"wavelen": "!OBS.wavelen"}
        #self.params.update(kwargs)

        self.wavelen = self.meta['wavelen']

        # field of view of the instrument
        # ..todo: get this from aperture list

        self.slicelist = self._file["Aperture List"].data
        #self.view = np.array([self.meta['naxis1'] * self.meta['pixscale'],
        #                      self.meta['nslice'] * self.meta['slicewidth']])
        self.view = np.array([self.slicelist['right'].max() -
                              self.slicelist['left'].min(),
                              self.slicelist['top'].max() -
                              self.slicelist['bottom'].min()])

        #for sli, spt in enumerate(self.spectral_traces.values()):
        #    spt.meta['xmin'] = self.slicelist['left'][sli]
        #    spt.meta['xmax'] = self.slicelist['right'][sli]
        #    spt.meta['ymin'] = self.slicelist['bottom'][sli]
        #    spt.meta['ymax'] = self.slicelist['top'][sli]

        #if self._file is not None:
        #    print(self._file)
        #    self.make_spectral_traces()

    def apply_to(self, obj, **kwargs):
        """
        overrides SpectralTraceList.apply_to()
        """
        if isinstance(obj, FOVSetupBase):
            # Create a single volume that covers the aperture and
            # the maximum wavelength range of LMS
            volumes = [self.spectral_traces[key].fov_grid()
                       for key in self.spectral_traces]
            wave_min = min([vol["wave_min"] for vol in volumes])
            wave_max = max([vol["wave_max"] for vol in volumes])
            extracted_vols = obj.extract(axes=["wave"],
                                         edges=([[wave_min, wave_max]]))
            obj.volumes = extracted_vols

        if isinstance(obj, FieldOfViewBase):
            # Application to field of view
            if obj.hdu is not None and obj.hdu.header['NAXIS'] == 3:
                obj.cube = obj.hdu
            elif obj.hdu is None and obj.cube is None:
                obj.cube = obj.make_cube_hdu()

            fovcube = obj.cube.data
            n_z, n_y, n_x = fovcube.shape
            fovwcs = WCS(obj.cube.header)
            # Make this linear to avoid jump at RA 0 deg
            fovwcs.wcs.ctype = ['LINEAR', 'LINEAR', fovwcs.wcs.ctype[2]]
            fovwcs_spat = fovwcs.sub(2)
            ny_slice = self.meta['slice_samples'] #

            # Spatial pixel coordinates
            xslice, yslice = np.meshgrid(np.arange(n_x),
                                         np.arange(ny_slice))

            fovimage = np.zeros((obj.detector_header['NAXIS2'],
                                 obj.detector_header['NAXIS1']),
                                dtype=np.float32)


            for sptid, spt in self.spectral_traces.items():
                ymin = spt.meta['fov']['y_min']
                ymax = spt.meta['fov']['y_max']

                slicewcs = deepcopy(fovwcs)

                slicewcs.wcs.ctype = ['LINEAR', 'LINEAR', slicewcs.wcs.ctype[2]]
                slicewcs.wcs.crpix[1] = (ny_slice + 1) / 2
                slicewcs.wcs.crval[1] = (ymin + ymax) / 2 / 3600
                slicewcs.wcs.cdelt[1] = (ymax - ymin) / ny_slice / 3600
                slicewcs_spat = slicewcs.sub(2)

                # World coordinates for the slice
                xworld, yworld = slicewcs_spat.all_pix2world(xslice,
                                                             yslice, 0)
                # FOV pixel coordinates for the slice
                xfov, yfov = fovwcs_spat.all_world2pix(xworld, yworld, 0)

                slicecube = np.zeros((n_z, ny_slice, n_x))
                for islice in range(n_z):
                    ifov = RectBivariateSpline(np.arange(n_y),
                                               np.arange(n_x),
                                               fovcube[islice], kx=1, ky=1)
                    slicecube[islice] = ifov(yfov, xfov, grid=False)

                slicefov = FieldOfView(obj.header,
                                       [obj.meta['wave_min'], obj.meta['wave_max']])
                slicefov.detector_header = obj.detector_header
                slicefov.meta['xi_min'] = obj.meta['xi_min']
                slicefov.meta['xi_max'] = obj.meta['xi_max']
                slicefov.meta['trace_id'] = sptid
                slicefov.cube = fits.ImageHDU(header=slicewcs.to_header(),
                                              data=slicecube)
                #slicefov.cube.writeto(f"slicefov_{sptid}.fits", overwrite=True)
                slicefov.hdu = spt.map_spectra_to_focal_plane(slicefov)

                sxmin = slicefov.hdu.header['XMIN']
                sxmax = slicefov.hdu.header['XMAX']
                symin = slicefov.hdu.header['YMIN']
                symax = slicefov.hdu.header['YMAX']
                fovimage[symin:symax, sxmin:sxmax] += slicefov.hdu.data

            obj.hdu = fits.ImageHDU(data=fovimage, header=obj.detector_header)

        return obj

    def make_spectral_traces(self):
        """
        Compute the transformations by interpolation
        """
        #nslice = len(self._file['Aperture List'].data)
        # determine echelle order and angle from specified wavelength
        tempres = self._angle_from_lambda()
        self.meta['order'] = tempres['Ord']
        self.meta['angle'] = tempres['Angle']

        spec_traces = {}
        for sli in np.arange(self.meta['nslice']):
            slicename = "Slice " + str(sli + 1)
            spec_traces[slicename] = MetisLMSSpectralTrace(
                self._file,
                spslice=sli, params=self.meta)

        self.spectral_traces = spec_traces

    def _angle_from_lambda(self):
        """
        Determine optimal echelle rotation angle for wavelength
        """
        lam = from_currsys(self.meta['wavelen'])
        grat_spacing = self.meta['grat_spacing']
        wcal = self._file['WCAL'].data
        return echelle_setting(lam, grat_spacing, wcal)


class MetisLMSSpectralTrace(SpectralTrace):
    """
    SpectralTrace for the METIS LM spectrograph
    """
    _class_params = {
        "naxis1": 122,
        "nslice": 28,
        "slicewidth": 0.0207,  # arcsec
        "pixscale": 0.0082,    # arcsec
        "grat_spacing": 18.2,
        "plate_scale": 0.303,
    }

    def __init__(self, hdulist, spslice, params, **kwargs):
        polyhdu = hdulist['Polynomial coefficients']
        params.update(kwargs)
        params['aperture_id'] = spslice
        params['slice'] = spslice
        super().__init__(polyhdu, **params)

        self._file = hdulist
        self.meta['description'] = "Slice " + str(spslice + 1)
        self.meta['trace_id'] = f"Slice {spslice + 1}"
        self.meta.update(params)
        # Provisional:
        self.meta['fov'] = self.fov_grid()

    def fov_grid(self):
        """
        Provide information on the source space volume required by the effect

        Returns
        -------
        A dictionary with entries `wave_min` and `wave_max`, `x_min`, `y_min`,
        `x_max`, `y_max`. Spatial limits refer to the sky and are given in
        arcsec.
        """

        aperture = self._file['Aperture list'].data[self.meta['slice']]
        x_min = aperture['left']
        x_max = aperture['right']
        y_min = aperture['bottom']
        y_max = aperture['top']
        trace_id = self.meta['trace_id']

        layout = ioascii.read(find_file("!DET.layout.file_name"))
        det_lims = {}
        xhw = layout['pixel_size'] * layout['x_size'] / 2
        yhw = layout['pixel_size'] * layout['y_size'] / 2
        det_lims['xd_min'] = min(layout['x_cen'] - xhw)
        det_lims['xd_max'] = max(layout['x_cen'] + xhw)
        det_lims['yd_min'] = min(layout['y_cen'] - yhw)
        det_lims['yd_max'] = max(layout['y_cen'] + yhw)
        wave_min, wave_max = self.get_waverange(det_lims)

        # ..todo: just a hack - xi and x are the same except xi is a quantity
        xi_min = quantify(x_min, u.arcsec)
        xi_max = quantify(x_max, u.arcsec)

        return {"x_min": x_min, "x_max": x_max,
                "y_min": y_min, "y_max": y_max,
                "xi_min": xi_min, "xi_max": xi_max,
                "wave_min": wave_min, "wave_max": wave_max,
                "trace_id": trace_id}

    def get_waverange(self, det_mm_lims):
        """Determine wavelength range that spectral trace covers on image plane"""
        xmin = det_mm_lims['xd_min']
        xmax = det_mm_lims['xd_max']

        lam0 = from_currsys(self.meta['wavelen'])
        xi0 = 0.
        ymid = self.xilam2y(xi0, lam0)[0]   # estimate y level of trace

        waverange = self.xy2lam(np.array([xmin, xmax]), np.array([ymid, ymid]),
                                grid=False)

        return waverange.min(), waverange.max()

    def compute_interpolation_functions(self):
        """
        Define the transforms between (xi, lam) and (x, y).
        The LMS transforms actually operate on phase rather than
        wavelength, hence the necessity of defining pre- and
        posttransforms on the lam variable.
        """
        matrices = self.get_matrices()
        # matrices are transposed to align argument sequence
        # with the name of the functions
        self.xilam2x = Transform2D(matrices['A'].T,
                                   pretransform_x=self.sky2fp,
                                   pretransform_y=self.lam2phase)
        self.xilam2y = Transform2D(matrices['B'].T,
                                   pretransform_x=self.sky2fp,
                                   pretransform_y=self.lam2phase)
        self.xy2lam = Transform2D(matrices['AI'],
                                  posttransform=self.phase2lam)
        self.xy2xi = Transform2D(matrices['BI'],
                                 posttransform=self.fp2sky)


    def get_matrices(self):
        """
        Extract matrix from lms_dist_poly.txt

        Evaluate polynomial to obtain matrices A, B, AI and BI at grism angle
        given echelle order and slice number

        Parameters
        ----------
        order : int
            Echelle order
        spslice : int
            Slice number
        angle : float
            Grism angle in degrees

        Returns
        -------
        dict of four np.arrays of shape (4, 4) each
        """
        spslice = self.meta['slice']
        order = self.meta['order']
        angle = self.meta['angle']
        matnames = ['A', 'B', 'AI', 'BI']
        matrices = {}

        poly = self.table
        for matid in range(4):
            select = ((poly['Ord'] == order) *
                      (poly['Sli'] == spslice) *
                      (poly['Mat'] == matid))
            if not np.any(select):
                raise KeyError("Combination of Order, Slice not found")

            subpoly = poly[select]
            thematrix = np.zeros((4, 4))
            for i in range(4):
                for j in range(4):
                    sel_ij = (subpoly['Row'] == i) * (subpoly['Col'] == j)
                    thematrix[i, j] = (subpoly['A11'][sel_ij] * angle**3 +
                                       subpoly['A12'][sel_ij] * angle**2 +
                                       subpoly['A21'][sel_ij] * angle +
                                       subpoly['A22'][sel_ij])
            matrices[matnames[matid]] = thematrix

        return matrices


    def lam2phase(self, lam):
        """
        Convert wavelength to phase

        Phase is lam * order / (2 * grat_spacing).

        Parameters
        ----------
        lam : ndarray (float)
            wavelength (um)

        Returns
        -------
        Phase : ndarray
        """
        return self.meta['order'] * lam / (2 * self.meta['grat_spacing'])

    def phase2lam(self, phase):
        """
        Convert phase to wavelength

        Wavelength is phase * 2 * grat_spacing / order

        Parameters
        ----------
        phase : ndarray (float)
            phase (dimensionless)

        Returns
        -------
        wavelength : ndarray (um)
        """
        return 2 * self.meta['grat_spacing'] * phase / self.meta['order']

    def sky2fp(self, xi):
        """
        Convert position in arcsec to position in FP2
        """
        return xi / self.meta['plate_scale']

    def fp2sky(self, fp_x):
        """
        Convert position in FP2 to position on sky
        """
        return fp_x * self.meta['plate_scale']


    def __repr__(self):
        msg = '<MetisLMSSpectralTrace> "{}" : {} um : Order {} : Angle {}'\
            ''.format(self.meta["description"],
                      from_currsys(self.meta["wavelen"]),
                      self.meta["order"],
                      self.meta["angle"])
        return msg


def echelle_setting(wavelength, grat_spacing, wcal_def):
    """
    Determine optimal echelle rotation angle for wavelength

    Parameters
    ----------
    lambda : float
            central wavelength in microns
    grat_spacing : float
            grating rule spacing in microns
    wcal_def: fits.TableHDU, fits.BinTableHDU, Table, str
            definition of the wavelength calibration parameters
            If str, interpreted as name of a fits file, with a
            table extension 'WCAL'.

    Returns
    -------
    a `dict` with entries
    - `Ord`: echelle order
    - `Angle`: grism angle
    - `Phase`: phase
    """
    if isinstance(wcal_def, fits.FITS_rec):
        wcal = wcal_def
    elif isinstance(wcal_def, (fits.TableHDU, fits.BinTableHDU)):
        # Read wcal extension of layout file
        wcal = wcal_def.data
    elif isinstance(wcal_def, Table):
        wcal = wcal_def
    elif isinstance(wcal_def, str):
        try:
            wcal = fits.getdata(wcal_def, extname='WCAL')
        except OSError:
            wcal = ioascii.read(wcal_def, comment="^#", format="csv")
    else:
        raise TypeError("wcal_def not in recognised format:", wcal_def)

    # Compute angles, determine which order gives angle closest to zero
    angles = wcal['c0'] * wavelength + wcal['c1']
    imin = np.argmin(np.abs(angles))

    # Extract parameters
    order = wcal['Ord'][imin]
    angle = angles[imin]

    # Compute the phase corresponding to the wavelength
    phase = wavelength * order / (2 * grat_spacing)

    return {"Ord": order, "Angle": angle, "Phase": phase}


class MetisLMSImageSlicer(ApertureMask):
    """
    Treats the METIS LMS image slicer as an aperture mask effect. This helps
    in building a FieldOfView object that combines the spatial field of the slicer
    with the spectral range covered by the LMS setting.

    The effect differs from its parent class `ApertureMask` in the initialisation
    from the `Aperture List` extension of the trace file `!OBS.trace_file`.
    """
    def __init__(self, filename, ext_id="Aperture List", **kwargs):
        filename = find_file(from_currsys(filename))
        ap_hdr = fits.getheader(filename, extname=ext_id)
        ap_list = fits.getdata(filename, extname=ext_id)
        xmin, xmax = ap_list['left'].min(), ap_list['right'].max()
        ymin, ymax = ap_list['bottom'].min(), ap_list['top'].max()
        slicer_dict = {"x": [xmin, xmax, xmax, xmin],
                       "y": [ymin, ymin, ymax, ymax]}
        try:
            kwargs["x_unit"] = ap_hdr['X_UNIT']
            kwargs["y_unit"] = ap_hdr['Y_UNIT']
        except KeyError:
            pass

        super().__init__(array_dict=slicer_dict, id="LMS slicer",
                         conserve_image=True, **kwargs)


class MetisLMSEfficiency(TERCurve):
    """
    Computes the grating efficiency (blaze function) for the METIS LMS.

    The procedure is described in E-REP-ATC-MET-1016_1.0. For a given order
    (determined by the central wavelength) the grating efficiency is modelled
    as a squared sinc function of wavelength via the grating angle.
    """
    _class_params = {
        "grat_spacing": 18.2,
        "eff_wid": 0.23,
        "eff_max": 0.75
    }

    def __init__(self, **kwargs):
        self.meta = self._class_params
        self.meta.update(kwargs)

        filename = find_file(self.meta['filename'])
        wcal = fits.getdata(filename, extname='WCAL')
        if 'wavelen' in kwargs:
            wavelen = from_currsys(kwargs['wavelen'])
            grat_spacing = self.meta['grat_spacing']
            ech = echelle_setting(wavelen, grat_spacing, wcal)
            self.meta['order'] = ech['Ord']
        else:
            wavelen = None

        lam, efficiency = self.make_ter_curve(wcal, wavelen)

        super().__init__(wavelength=lam,
                         transmission=efficiency,
                         emissivity=np.zeros_like(lam),
                         **self.meta)

    def make_ter_curve(self, wcal, wavelen=None):
        """Compute the blaze function for the selected order"""
        order = self.meta['order']
        eff_wid = self.meta['eff_wid']
        eff_max = self.meta['eff_max']

        wcal_ord = wcal[wcal['Ord'] == self.meta['order']]

        if wavelen is not None:
            lam = np.linspace(wavelen - 0.2, wavelen + 0.2, 1001)
            angle = wcal_ord['c0'] * lam + wcal_ord['c1']
        else:
            angle = np.linspace(7, -7, 10001)
            lam = wcal_ord['ic0'] * angle + wcal_ord['ic1']

        phase = order * np.pi * np.sin(np.deg2rad(angle)) * eff_wid
        efficiency = eff_max * np.sinc(phase / np.pi)**2

        return lam, efficiency



import logging

import numpy as np
from astropy import units as u
from synphot import SourceSpectrum, BlackBody1D, Empirical1D

from ..utils import quantify, extract_type_from_unit, extract_base_from_unit


def make_emission_from_emissivity(temp, emiss_src_spec):
    """
    Create an emission SourceSpectrum using a blackbody and an emissivity curve

    Parameters
    ----------
    temp : float, Quantity
        [Kelvin] If float, then must be in Kelvin
    emiss_src_spec : synphot.SpectralElement
        An emissivity response curve in the range [0..1]

    Returns
    -------
    flux : synphot.SourceSpectrum

    """

    if isinstance(temp, u.Quantity):
        temp = temp.to(u.Kelvin, equivalencies=u.temperature()).value

    if emiss_src_spec is None:
        logging.warning("Either emission or emissivity must be set")
        flux = None
    else:
        flux = SourceSpectrum(BlackBody1D, temperature=temp)
        flux.meta["solid_angle"] = u.sr**-1
        flux = flux * emiss_src_spec
        flux.meta["history"] = ["Created from Blackbody curve. Units are to be"
                                " understood as per steradian"]

    return flux


def make_emission_from_array(flux, wave, meta):
    """
    Create an emission SourceSpectrum using array.

    Takes care of bins and solid angles. The solid_angle is kept in the returned
    SourceSpectrum meta dictionary under self.meta["solid_angle"]

    Parameters
    ----------
    flux : array-like, Quantity
        if flux is not an array, the ``emission_unit`` must be in meta dict
    wave : array-like, Quantity
        if flux is not an array, the ``wavelength_unit`` must be in meta dict
    meta : dict

    Returns
    -------
    flux : synphot.SourceSpectrum

    """

    if not isinstance(flux, u.Quantity):
        if "emission_unit" in meta:
            flux = quantify(flux, meta["emission_unit"])
        else:
            logging.warning("emission_unit must be set in self.meta, "
                          "or emission must be an astropy.Quantity")
            flux = None

    if isinstance(wave, u.Quantity) and isinstance(flux, u.Quantity):
        flux_unit, angle = extract_type_from_unit(flux.unit, "solid angle")
        flux = flux / angle

        if is_flux_binned(flux.unit):
            flux = normalise_binned_flux(flux, wave)

        orig_unit = flux.unit
        flux = SourceSpectrum(Empirical1D, points=wave,
                              lookup_table=flux)
        flux.meta["solid_angle"] = angle
        flux.meta["history"] = ["Created from emission array with units {}"
                                "".format(orig_unit)]
    else:
        logging.warning("wavelength and emission must be "
                      "astropy.Quantity py_objects")
        flux = None

    return flux


def normalise_binned_flux(flux, wave):
    """
    Convert a binned flux Quantity array back into flux density

    The flux density normalising unit is taken from the wavelength Quantity unit

    Parameters
    ----------
    flux : array-like Quantity
        flux unit must include ``bin``, e.g. ``ph s-1 m-2 bin-1``
    wave : array-like Quantity

    Returns
    -------
    flux : array-like Quantity

    """

    bins = np.zeros(len(wave)) * wave.unit
    bins[:-1] = 0.5 * np.diff(wave)
    bins[1:] += 0.5 * np.diff(wave)
    # bins[0] *= 2.   # edge bins only have half the flux of other bins
    # bins[-1] *= 2.

    bin_unit = extract_base_from_unit(flux.unit, u.bin)[1]
    flux = flux / bins / bin_unit

    return flux


def is_flux_binned(unit):
    """
    Checks if the (flux) unit is a binned unit

    Parameters
    ----------
    unit : Unit

    Returns
    -------
    flag : bool

    """
    unit = unit**1
    flag = False
    # unit.physical_type is a string in astropy<=4.2 and a PhysicalType
    # class in astropy==4.3 and thus has to be cast to a string first.
    if u.bin in unit._bases or "flux density" not in str(unit.physical_type):
        flag = True

    return flag



"""
Electronic detector effects - related to detector readout

Classes:
- DetectorModePropertiesSetter - set parameters for readout mode
- AutoExposure - determine DIT and NDIT automatically
- SummedExposure - simulates a summed stack of ``ndit`` exposures
- PoorMansHxRGReadoutNoise - simple readout noise for HAWAII detectors
- BasicReadoutNoise - readout noise
- ShotNoise - realisation of Poissonian photon noise
- DarkCurrent - add dark current
- LinearityCurve - apply detector (non-)linearity and saturation
- ReferencePixelBorder
- BinnedImage
- Bias - adds constant bias level to readout

Functions:
- make_ron_frame
- pseudo_random_field
"""
import logging

import numpy as np
import matplotlib.pyplot as plt

from astropy.io import fits

from .. import rc
from . import Effect
from ..base_classes import DetectorBase, ImagePlaneBase
from ..utils import from_currsys
from .. import utils


class DetectorModePropertiesSetter(Effect):
    """
    Sets mode specific curr_sys properties for different detector readout modes

    A little class (``DetectorModePropertiesSetter``) that allows different
    ``"!DET"`` properties to be set on the fly.

    Parameters
    ----------
    mode_properties : dict
        A dictionary containing the DET parameters to be changed for each mode.
        See below for an example yaml entry.

    Examples
    --------
    Add the values for the different detector readout modes to all the relevant
    detector yaml files. In this case the METIS HAWAII (L, M band) and GeoSnap
    (N band) detectors: METIS_DET_IMG_LM.yaml , METIS_DET_IMG_N.yaml
    ::

        - name: lm_detector_readout_parameters
          class: DetectorModePropertiesSetter
          kwargs:
            mode_properties:
              fast:
                mindit: 0.04
                full_well: !!float 1e5
                ron: 70
              slow:
                mindit: 1.3
                full_well: !!float 1e5
                ron: 14

    Add the OBS dict entry !OBS.detector_readout_mode to the properties section
    of the mode_yamls descriptions in the default.yaml files.
    ::

        mode_yamls:
          - object: observation
            alias: OBS
            name: lss_l
            yamls:
              ...
            properties:
              ...
              detector_readout_mode: slow

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        params = {"z_order": [299, 900]}
        self.meta.update(params)
        self.meta.update(kwargs)

        required_keys = ['mode_properties']
        utils.check_keys(self.meta, required_keys, action="error")

        self.mode_properties = kwargs['mode_properties']

    def apply_to(self, obj, **kwargs):
        mode_name = kwargs.get('detector_readout_mode',
                               from_currsys("!OBS.detector_readout_mode"))
        if isinstance(obj, ImagePlaneBase) and mode_name == "auto":
            mode_name = self.select_mode(obj, **kwargs)
            print("Detector mode set to", mode_name)

        self.meta['detector_readout_mode'] = mode_name
        props_dict = self.mode_properties[mode_name]
        rc.__currsys__["!OBS.detector_readout_mode"] = mode_name
        for key, value in props_dict.items():
            rc.__currsys__[key] = value

        return obj

    def list_modes(self):
        """Return list of available detector modes"""
        return utils.pretty_print_dict(self.mode_properties)

    def select_mode(self, obj, **kwargs):
        """Automatically select detector mode based on image plane peak value

        Select the mode with lowest readnoise that does not saturate the detector.
        When all modes saturate, select the mode with the lowest saturation level
        (peak to full_well).
        """
        immax = np.max(obj.data)
        fillfrac = kwargs.get("fill_frac",
                              from_currsys("!OBS.auto_exposure.fill_frac"))

        goodmodes = []
        goodron = []
        badmodes = []
        satlevel = []
        for modeid, modeprops in self.mode_properties.items():
            mindit = modeprops["!DET.mindit"]
            fullwell = modeprops["!DET.full_well"]
            if immax * mindit < fillfrac * fullwell:
                goodmodes.append(modeid)
                goodron.append(modeprops["!DET.readout_noise"])
            else:
                badmodes.append(modeid)
                satlevel.append(immax * mindit / fullwell)

        if not goodmodes:
            return badmodes[np.argmin(satlevel)]

        return goodmodes[np.argmin(goodron)]


class AutoExposure(Effect):
    """
    Determine DIT and NDIT automatically from ImagePlane

    DIT is determined such that the maximum value in the incident photon flux
    (including astronomical source, sky and thermal backgrounds) fills
    the full well of the detector (``!DET.full_well``) to a given fraction
    (``!OBS.autoexposure.fill_frac``). NDIT is determined such that
    ``DIT`` * ``NDIT`` results in the requested exposure time.

    The requested exposure time is taken from ``!OBS.exptime``.

    The effects sets the parameters `!OBS.dit` and `!OBS.ndit`.

    Examples
    --------
    The parameters `!OBS.exptime`, `!DET.full_well` and `!DET.mindit` should
    be defined as properties in the respective subsections.
    ::

       name: auto_exposure
       description: automatic determination of DIT and NDIT
       class: AutoExposure
       include: True
       kwargs:
           fill_frac: "!OBS.auto_exposure.fill_frac"

    """
    def __init__(self, **kwargs):
        """
        The effect is the first detector effect, hence essentially operates
        on the `ImagePlane`, mapped to the detector array.
        """
        super().__init__(**kwargs)
        params = {"z_order": [902]}
        self.meta.update(params)
        self.meta.update(kwargs)

        required_keys = ['fill_frac', 'full_well', 'mindit']
        utils.check_keys(self.meta, required_keys, action="error")

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, (ImagePlaneBase, DetectorBase)):
            implane_max = np.max(obj.data)
            exptime = kwargs.get('exptime', from_currsys("!OBS.exptime"))
            mindit = from_currsys(self.meta["mindit"])

            if exptime is None:
                exptime = from_currsys("!OBS.dit") * from_currsys("!OBS.ndit")
            print(f"Requested exposure time: {exptime:.3f} s")

            if exptime < mindit:
                print(f"    increased to MINDIT: {mindit:.3f} s")
                exptime = mindit

            full_well = from_currsys(self.meta["full_well"])
            fill_frac = kwargs.get("fill_frac",
                                   from_currsys(self.meta["fill_frac"]))
            dit = fill_frac * full_well / implane_max

            # np.ceil so that dit is at most what is required for fill_frac
            ndit = int(np.ceil(exptime / dit))
            dit = exptime / ndit

            # dit must be at least mindit, this might lead to saturation
            # ndit changed so that exptime is not exceeded (hence np.floor)
            if dit < from_currsys(self.meta["mindit"]):
                dit = from_currsys(self.meta["mindit"])
                ndit = int(np.floor(exptime / dit))
                print("Warning: The detector will be saturated!")
                # ..todo: turn into proper warning

            print("Exposure parameters:")
            print(f"                DIT: {dit:.3f} s  NDIT: {ndit}")
            print(f"Total exposure time: {dit * ndit:.3f} s")

            rc.__currsys__['!OBS.dit'] = dit
            rc.__currsys__['!OBS.ndit'] = ndit

        return obj


class SummedExposure(Effect):
    """
    Simulates a summed stack of ``ndit`` exposures

    """
    def __init__(self, **kwargs):
        super(SummedExposure, self).__init__(**kwargs)
        params = {"z_order": [860]}
        self.meta.update(params)
        self.meta.update(kwargs)

        required_keys = ["dit", "ndit"]
        utils.check_keys(self.meta, required_keys, action="error")

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, DetectorBase):
            dit = from_currsys(self.meta["dit"])
            ndit = from_currsys(self.meta["ndit"])

            obj._hdu.data *= dit * ndit

        return obj


class Bias(Effect):
    """
    Adds a constant bias level to readout

    """
    def __init__(self, **kwargs):
        super(Bias, self).__init__(**kwargs)
        params = {"z_order": [855]}
        self.meta.update(params)
        self.meta.update(kwargs)

        required_keys = ["bias"]
        utils.check_keys(self.meta, required_keys, action="error")

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, DetectorBase):
            biaslevel = from_currsys(self.meta["bias"])
            obj._hdu.data += biaslevel

        return obj

class PoorMansHxRGReadoutNoise(Effect):
    def __init__(self, **kwargs):
        super(PoorMansHxRGReadoutNoise, self).__init__(**kwargs)
        params = {"z_order": [811],
                  "pedestal_fraction": 0.3,
                  "read_fraction": 0.4,
                  "line_fraction": 0.25,
                  "channel_fraction": 0.05,
                  "random_seed": "!SIM.random.seed",
                  "report_plot_include": False,
                  "report_table_include": False}
        self.meta.update(params)
        self.meta.update(kwargs)

        self.required_keys = ["noise_std", "n_channels", "ndit"]
        utils.check_keys(self.meta, self.required_keys, action="error")

    def apply_to(self, det, **kwargs):
        if isinstance(det, DetectorBase):
            self.meta["random_seed"] = from_currsys(self.meta["random_seed"])
            if self.meta["random_seed"] is not None:
                np.random.seed(self.meta["random_seed"])

            from_currsys(self.meta)
            ron_keys = ["noise_std", "n_channels", "channel_fraction",
                        "line_fraction", "pedestal_fraction", "read_fraction"]
            ron_kwargs = {key: self.meta[key] for key in ron_keys}
            ron_kwargs["image_shape"] = det._hdu.data.shape

            ron_frame = make_ron_frame(**ron_kwargs)
            stacked_ron_frame = np.zeros_like(ron_frame)
            for i in range(self.meta["ndit"]):
                dx = np.random.randint(0, ron_frame.shape[1])
                dy = np.random.randint(0, ron_frame.shape[0])
                stacked_ron_frame += np.roll(ron_frame, (dy, dx), axis=(0, 1))

            # .. todo: this .T is ugly. Work out where things are getting switched and remove it!
            det._hdu.data += stacked_ron_frame.T

        return det

    def plot(self, det, new_figure=False, **kwargs):
        dtcr = self.apply_to(det)
        plt.imshow(dtcr.data, origin="lower")

    def plot_hist(self, det, **kwargs):
        dtcr = self.apply_to(det)
        plt.hist(dtcr.data.flatten())


class BasicReadoutNoise(Effect):
    """Readout noise computed as: ron * sqrt(NDIT)"""
    def __init__(self, **kwargs):
        super(BasicReadoutNoise, self).__init__(**kwargs)
        self.meta["z_order"] = [811]
        self.meta["random_seed"] = "!SIM.random.seed"
        self.meta.update(kwargs)

        self.required_keys = ["noise_std", "ndit"]
        utils.check_keys(self.meta, self.required_keys, action="error")

    def apply_to(self, det, **kwargs):
        if isinstance(det, DetectorBase):
            ndit = from_currsys(self.meta["ndit"])
            ron = from_currsys(self.meta["noise_std"])
            noise_std = ron * np.sqrt(float(ndit))

            random_seed = from_currsys(self.meta["random_seed"])
            if random_seed is not None:
                np.random.seed(random_seed)
            det._hdu.data += np.random.normal(loc=0, scale=noise_std,
                                              size=det._hdu.data.shape)

        return det

    def plot(self, det):
        dtcr = self.apply_to(det)
        plt.imshow(dtcr.data)

    def plot_hist(self, det, **kwargs):
        dtcr = self.apply_to(det)
        plt.hist(dtcr.data.flatten())


class ShotNoise(Effect):
    def __init__(self, **kwargs):
        super(ShotNoise, self).__init__(**kwargs)
        self.meta["z_order"] = [820]
        self.meta["random_seed"] = "!SIM.random.seed"
        self.meta.update(kwargs)

    def apply_to(self, det, **kwargs):
        if isinstance(det, DetectorBase):
            self.meta["random_seed"] = from_currsys(self.meta["random_seed"])
            if self.meta["random_seed"] is not None:
                np.random.seed(self.meta["random_seed"])

            # ! poisson(x) === normal(mu=x, sigma=x**0.5)
            # Windows has a problem with generating poisson values above 2**30
            # Above ~100 counts the poisson and normal distribution are
            # basically the same. For large arrays the normal distribution
            # takes only 60% as long as the poisson distribution
            data = det._hdu.data

            # Check if there are negative values in the data
            negvals = np.sum(data < 0)
            if negvals:
                logging.warning(f"Effect ShotNoise: {negvals} negative pixels")
                data[data < 0] = 0

            below = data < 2**20
            above = np.invert(below)
            data[below] = np.random.poisson(data[below]).astype(float)
            data[above] = np.random.normal(data[above], np.sqrt(data[above]))
            data = np.floor(data)
            new_imagehdu = fits.ImageHDU(data=data, header=det._hdu.header)
            det._hdu = new_imagehdu

        return det

    def plot(self, det):
        dtcr = self.apply_to(det)
        plt.imshow(dtcr.data)

    def plot_hist(self, det, **kwargs):
        dtcr = self.apply_to(det)
        plt.hist(dtcr.data.flatten())


class DarkCurrent(Effect):
    """
    required: dit, ndit, value
    """
    def __init__(self, **kwargs):
        super(DarkCurrent, self).__init__(**kwargs)
        self.meta["z_order"] = [830]

        required_keys = ["value", "dit", "ndit"]
        utils.check_keys(self.meta, required_keys, action="error")

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, DetectorBase):
            if isinstance(from_currsys(self.meta["value"]), dict):
                dtcr_id = obj.meta[utils.real_colname("id", obj.meta)]
                dark = from_currsys(self.meta["value"][dtcr_id])
            elif isinstance(from_currsys(self.meta["value"]), float):
                dark = from_currsys(self.meta["value"])
            else:
                raise ValueError("<DarkCurrent>.meta['value'] must be either"
                                 "dict or float: {}".format(self.meta["value"]))

            dit = from_currsys(self.meta["dit"])
            ndit = from_currsys(self.meta["ndit"])

            obj._hdu.data += dark * dit * ndit

        return obj

    def plot(self, det, **kwargs):
        dit = from_currsys(self.meta["dit"])
        ndit = from_currsys(self.meta["ndit"])
        total_time = dit * ndit
        times = np.linspace(0, 2*total_time, 10)
        dtcr = self.apply_to(det)
        dark_level = dtcr.data[0, 0] / total_time  # just read one pixel
        levels = dark_level * times
        plt.plot(times, levels, **kwargs)
        plt.xlabel("time")
        plt.ylabel("dark level")


class LinearityCurve(Effect):
    """
    Detector linearity effect

    The detector linearity curve is set in terms of `incident` flux (e/s) and `measured`
    detector values (ADU).

    Examples
    --------

    The effect can be instantiated in various ways.::

        - name: detector_linearity
          class: LinearityCurve
          kwargs:
            file_name: FPA_linearity.dat

        - name: detector_linearity
          class: LinearityCurve
          kwargs:
            array_dict: {incident: [0, 77000, 999999999999],
                         measured: [0, 77000, 77000]}

        - name: detector_linearity
          class: LinearityCurve
          kwargs:
            incident: [0, 77000, 99999999]
            measured: [0, 77000, 77000]

    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        params = {"z_order": [840],
                  "report_plot_include": True,
                  "report_table_include": False}
        self.meta.update(params)
        self.meta.update(kwargs)

        self.required_keys = ["ndit"]
        utils.check_keys(self.meta, self.required_keys, action="error")

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, DetectorBase):
            ndit = from_currsys(self.meta["ndit"])
            if self.table is not None:
                incident = self.table["incident"] * ndit
                measured = self.table["measured"] * ndit
            else:
                incident = np.asarray(from_currsys(self.meta["incident"])) * ndit
                measured = np.asarray(from_currsys(self.meta["measured"])) * ndit
            obj._hdu.data = np.interp(obj._hdu.data, incident, measured)

        return obj

    def plot(self, **kwargs):
        plt.gcf().clf()

        ndit = from_currsys(self.meta["ndit"])
        incident = self.table["incident"] * ndit
        measured = self.table["measured"] * ndit

        plt.loglog(incident, measured, **kwargs)
        plt.xlabel("Incident [ph s$^-1$]")
        plt.ylabel("Measured [e- s$^-1$]")

        return plt.gcf()


class ReferencePixelBorder(Effect):
    def __init__(self, **kwargs):
        super(ReferencePixelBorder, self).__init__(**kwargs)
        self.meta["z_order"] = [780]
        val = 0
        if "all" in kwargs:
            val = int(kwargs["all"])
        widths = {key: val for key in ["top", "bottom", "left", "right"]}
        self.meta.update(widths)
        self.meta.update(kwargs)

    def apply_to(self, implane, **kwargs):
        # .. todo: should this be ImagePlaneBase here?
        if isinstance(implane, ImagePlaneBase):
            if self.meta["top"] > 0:
                implane.hdu.data[:, -self.meta["top"]:] = 0
            if self.meta["bottom"] > 0:
                implane.hdu.data[:, :self.meta["bottom"]] = 0
            if self.meta["right"] > 0:
                implane.hdu.data[-self.meta["right"]:, :] = 0
            if self.meta["left"] > 0:
                implane.hdu.data[:self.meta["left"], :] = 0

        return implane

    def plot(self, implane, **kwargs):
        implane = self.apply_to(implane)
        plt.imshow(implane.data, origin="bottom", **kwargs)
        plt.show()


class BinnedImage(Effect):
    def __init__(self, **kwargs):
        super(BinnedImage, self).__init__(**kwargs)
        self.meta["z_order"] = [870]

        self.required_keys = ["bin_size"]
        utils.check_keys(self.meta, self.required_keys, action="error")

    def apply_to(self, det, **kwargs):
        if isinstance(det, DetectorBase):
            bs = from_currsys(self.meta["bin_size"])
            image = det._hdu.data
            h, w = image.shape
            new_image = image.reshape((h//bs, bs, w//bs, bs))
            det._hdu.data = new_image.sum(axis=3).sum(axis=1)

        return det


################################################################################


def make_ron_frame(image_shape, noise_std, n_channels, channel_fraction,
                   line_fraction, pedestal_fraction, read_fraction):
    shape = image_shape
    w_chan = max(1, shape[0] // n_channels)

    pixel_std = noise_std * (pedestal_fraction + read_fraction)**0.5
    line_std = noise_std * line_fraction**0.5
    if shape < (1024, 1024):
        pixel = np.random.normal(loc=0, scale=pixel_std, size=shape)
        line = np.random.normal(loc=0, scale=line_std, size=shape[1])
    else:
        pixel = pseudo_random_field(scale=pixel_std, size=shape)
        line = pixel[0]

    channel_std = noise_std * channel_fraction**0.5
    channel = np.repeat(np.random.normal(loc=0, scale=channel_std,
                                         size=n_channels), w_chan + 1, axis=0)

    ron_frame = (pixel + line).T + channel[:shape[0]]

    return ron_frame


def pseudo_random_field(scale=1, size=(1024, 1024)):
    n = 256
    image = np.zeros(size)
    batch = np.random.normal(loc=0, scale=scale, size=(2*n, 2*n))
    for y in range(0, size[1], n):
        for x in range(0, size[0], n):
            i, j = np.random.randint(n, size=2)
            dx, dy = min(size[0]-x, n), min(size[1]-y, n)
            image[x:x+dx, y:y+dy] = batch[i:i+dx, j:j+dy]

    return image



"""The Instrument data simulator for MICADO on the E-ELT"""

################################################################################
#                            TURN OFF WARNINGS                                 #
################################################################################
import sys
import logging
import warnings
import yaml
from astropy.utils.exceptions import AstropyWarning

warnings.simplefilter('ignore', UserWarning)
warnings.simplefilter('ignore', FutureWarning)
warnings.simplefilter('ignore', RuntimeWarning)     # warnings for the developer
warnings.simplefilter('ignore', category=AstropyWarning)
yaml.warnings({'YAMLLoadWarning': False})

################################################################################
#                         PACKAGE GLOBAL VARIABLES                             #
################################################################################

from . import rc

################################################################################
#                         SET BASIC LOGGING LEVEL                              #
################################################################################

root = logging.getLogger()
root.setLevel("DEBUG")            # DEBUG

if rc.__config__["!SIM.logging.log_to_file"] is True:
    file_path = rc.__config__["!SIM.logging.file_path"]
    write_mode = rc.__config__["!SIM.logging.file_open_mode"]
    file_handler = logging.FileHandler(file_path, write_mode)
    file_handler.setLevel(rc.__config__["!SIM.logging.file_level"])  # DEBUG
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)
    root.addHandler(file_handler)

if rc.__config__["!SIM.logging.log_to_console"] is True:
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.setLevel(rc.__config__["!SIM.logging.console_level"])  # WARNING
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    stdout_handler.setFormatter(formatter)
    root.addHandler(stdout_handler)


################################################################################
#                         IMPORT PACKAGE MODULES                               #
################################################################################

# Import all the modules to go under ScopeSim

from . import commands
from . import source
from . import optics
from . import detector
from . import effects
from . import server
from . import utils

# import specific classes from the modules to included in the global namespace

from .utils import bug_report
from .optics.optical_train import OpticalTrain
from .commands.user_commands import UserCommands
from .source.source import Source

from .server.database import (list_packages, download_packages, download_package,
                              list_example_data, download_example_data)

from .tests.mocks.load_basic_instrument import load_example_optical_train

################################################################################
#                          VERSION INFORMATION                                 #
################################################################################

try:
    from .version import version as __version__
except ImportError:
    __version__ = "Version number is not available"



"""Effects related to field masks, including spectroscopic slits"""
from os import path as pth
from copy import deepcopy
import logging
import yaml

import numpy as np
from matplotlib.path import Path
from astropy.io import fits
from astropy import units as u
from astropy.table import Table

from .effects import Effect
from ..optics import image_plane_utils as imp_utils
from ..base_classes import FOVSetupBase

from ..utils import quantify, quantity_from_table, from_currsys, check_keys


class ApertureMask(Effect):
    """
    Only provides the on-sky window coords of the Aperture

    - Case: Imaging
        - Covers the whole FOV of the detector
        - Round (with mask), square (without mask)
    - Case : LS Spec
        - Covers the slit FOV
        - Polygonal (with mask), square (without mask)
    - Case : IFU Spec
        - Covers the on-sky FOV of one slice of the IFU
        - Square (without mask)
    - Case : MOS Spec
        - Covers a single MOS fibre FOV
        - Round, Polygonal (with mask), square (without mask)

    The geometry of an ``ApertureMask`` can be initialised with the standard
    DataContainer methods (see Parameters below). Regardless of which method
    is used, the following columns must be present::

        x       y
        arcsec  arcsec
        float   float

    Certain keywords need to also be included in the ascii header::

        # id: <int>
        # conserve_image: <bool>
        # x_unit: <str>
        # y_unit: <str>

    If ``conserve_image`` is ``False``, the flux from all sources in the
    aperture is summed and distributed uniformly over the aperture area.


    Parameters
    ----------
    filename : str
        Path to ASCII file containing the columns listed above

    table : astropy.Table
        An astropy Table containing the columns listed above

    array_dict : dict
        A dictionary containing the columns listed above:
        ``{x: [...], y: [...], id: <int>, conserve_image: <bool>}``

    Other Parameters
    ----------------
    pixel_scale : float
        [arcsec] Defaults to ``"!INST.pixel_scale"`` from the config

    id : int
        An integer to identify the ``ApertureMask`` in a list of apertures

    """
    def __init__(self, **kwargs):
        if not np.any([key in kwargs for key in ["filename", "table",
                                                 "array_dict"]]):
            if "width" in kwargs and "height" in kwargs and \
                    "filename_format" in kwargs:
                kwargs = from_currsys(kwargs)
                w, h = kwargs["width"], kwargs["height"]
                kwargs["filename"] = kwargs["filename_format"].format(w, h)

        super(ApertureMask, self).__init__(**kwargs)
        params = {"pixel_scale": "!INST.pixel_scale",
                  "no_mask": True,
                  "angle": 0,
                  "shape": "rect",
                  "conserve_image": True,
                  "id": 0,
                  "report_plot_include": False,
                  "report_table_include": True,
                  "report_table_rounding": 4}

        self.meta["z_order"] = [80, 280, 380]
        self.meta.update(params)
        self.meta.update(kwargs)

        self._header = None
        self._mask = None
        self.mask_sum = None

        self.required_keys = ["filename", "table", "array_dict"]
        check_keys(kwargs, self.required_keys, "warning", all_any="any")

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, FOVSetupBase):
            x = quantity_from_table("x", self.table, u.arcsec).to(u.arcsec).value
            y = quantity_from_table("y", self.table, u.arcsec).to(u.arcsec).value
            obj.shrink(["x", "y"], ([min(x), max(x)], [min(y), max(y)]))

            # ..todo: HUGE HACK - Get rid of this!
            for vol in obj.volumes:
                vol["meta"]["xi_min"] = min(x) * u.arcsec
                vol["meta"]["xi_max"] = max(x) * u.arcsec

        return obj

    # Outdated. Remove when removing all old FOVManager code from effects
    def fov_grid(self, which="edges", **kwargs):
        """ Returns a header with the sky coordinates """
        logging.warning("DetectorList.fov_grid will be depreciated in v1.0")
        if which == "edges":
            self.meta.update(kwargs)
            return self.header
        elif which == "masks":
            self.meta.update(kwargs)
            return self.mask

    @property
    def hdu(self):
        return fits.ImageHDU(data=self.mask, header=self.header)

    @property
    def header(self):
        if not isinstance(self._header, fits.Header) \
                and "x" in self.table.colnames and "y" in self.table.colnames:
            self._header = self.get_header()
        return self._header

    def get_header(self):
        self.meta = from_currsys(self.meta)
        x = quantity_from_table("x", self.table, u.arcsec).to(u.deg).value
        y = quantity_from_table("y", self.table, u.arcsec).to(u.deg).value
        pix_scale_deg = self.meta["pixel_scale"] / 3600.
        header = imp_utils.header_from_list_of_xy(x, y, pix_scale_deg)
        header["APERTURE"] = self.meta["id"]
        header["ROT"] = self.meta["angle"]
        header["IMG_CONS"] = self.meta["conserve_image"]

        return header

    @property
    def mask(self):
        if not isinstance(self._header, fits.Header) \
                and "x" in self.table.colnames and "y" in self.table.colnames:
            self._mask = self.get_mask()
        return self._mask

    def get_mask(self):
        """
        For placing over FOVs if the Aperture is rotated w.r.t. the field
        """
        self.meta = from_currsys(self.meta)

        if self.meta["no_mask"] is False:
            x = quantity_from_table("x", self.table, u.arcsec).to(u.deg).value
            y = quantity_from_table("y", self.table, u.arcsec).to(u.deg).value
            pixel_scale_deg = self.meta["pixel_scale"] / 3600.
            mask = mask_from_coords(x, y, pixel_scale_deg)
        else:
            mask = None

        return mask

    def plot(self, new_figure=True):
        import matplotlib.pyplot as plt
        if new_figure:
            plt.gcf().clf()

        x = list(self.table["x"].data)
        y = list(self.table["y"].data)
        plt.plot(x + [x[0]], y + [y[0]])
        plt.gca().set_aspect("equal")

        return plt.gcf()


class RectangularApertureMask(ApertureMask):
    def __init__(self, **kwargs):
        super(RectangularApertureMask, self).__init__(**kwargs)
        params = {"x_unit" : "arcsec",
                  "y_unit" : "arcsec"}
        self.meta.update(params)
        self.meta.update(kwargs)
        check_keys(self.meta, ["x", "y", "width", "height"])

        self.table = self.get_table(**kwargs)

    # def fov_grid(self, which="edges", **kwargs):
    #     """ Returns a header with the sky coordinates """
    #     if which == "edges":
    #         self.table = self.get_table(**kwargs)
    #         return self.header      # from base class ApertureMask
    #
    #     elif which == "masks":
    #         self.meta.update(kwargs)
    #         return self.mask

    def get_table(self, **kwargs):
        self.meta.update(kwargs)
        x = from_currsys(self.meta["x"])
        y = from_currsys(self.meta["y"])
        dx = 0.5 * from_currsys(self.meta["width"])
        dy = 0.5 * from_currsys(self.meta["height"])
        xs = [x - dx, x + dx, x + dx, x - dx]
        ys = [y - dy, y - dy, y + dy, y + dy]
        tbl = Table(names=["x", "y"], data=[xs, ys], meta=self.meta)

        return tbl


class ApertureList(Effect):
    """
    A list of apertures, useful for IFU or MOS instruments

    Parameters
    ----------

    Examples
    --------

    File format
    -----------

    Much like an ApertureMask, an ApertureList can be initialised by either
    of the three standard DataContainer methods. The easiest is however to
    make an ASCII file with the following columns::

        id   left    right   top     bottom  angle  conserve_image  shape
             arcsec  arcsec  arcsec  arcsec  deg
        int  float   float   float   float   float  bool            str/int

    Acceptable ``shape`` entries are: ``round``, ``rect``, ``hex``, ``oct``, or
    an integer describing the number of corners the polygon should have.

    A polygonal mask is generated for a given ``shape`` and will be scaled
    to fit inside the edges of each aperture list row. The corners of each
    aperture defined by shape are found by finding equidistant positions around
    an ellipse constrained by the edges (``left``, ..., etc). An additional
    optional column ``offset`` may be added. This column describes the offset
    from 0 deg to the angle where the first corner is set.

    Additionally, the filename of an ``ApertureMask`` polygon file can be given.
    The geometry of the polygon defined in the file will be scaled to fit
    inside the edges of the row.

    .. note:: ``shape`` values ``"rect"`` and ``4`` do not produce equal results

       Both use 4 equidistant points around an ellipse constrained by
       [``left``, ..., etc]. However ``"rect"`` aims to fill the contraining
       area, while ``4`` simply uses 4 points on the ellipse.
       Consequently, ``4`` results in a diamond shaped mask covering only
       half of the constraining area filled by ``"rect"``.


    """
    def __init__(self, **kwargs):
        super(ApertureList, self).__init__(**kwargs)
        params = {"pixel_scale": "!INST.pixel_scale",
                  "n_round_corners": 32,        # number of corners use to estimate ellipse
                  "no_mask": False,             # .. todo:: is this necessary when we have conserve_image?
                  "report_plot_include": True,
                  "report_table_include": True,
                  "report_table_rounding": 4}
        self.meta["z_order"] = [81, 281]
        self.meta.update(params)
        self.meta.update(kwargs)

        if self.table is not None:
            required_keys = ["id", "left", "right", "top", "bottom", "angle",
                             "conserve_image", "shape"]
            check_keys(self.table.colnames, required_keys)

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, FOVSetupBase):
            new_vols = []
            for row in self.table:
                vols = obj.extract(["x", "y"], ([row["left"], row["right"]],
                                                [row["bottom"], row["top"] ]))
                for vol in vols:
                    vol["meta"]["aperture_id"] = row["id"]

                    # ..todo: HUGE HACK - Get rid of this!
                    vol["meta"]["xi_min"] = row["left"] * u.arcsec
                    vol["meta"]["xi_max"] = row["right"] * u.arcsec
                    vol["conserve_image"] = row["conserve_image"]
                    vol["shape"] = row["shape"]
                    vol["angle"] = row["angle"]

                new_vols += vols

            obj.volumes = new_vols

        return obj

    # def fov_grid(self, which="edges", **kwargs):
    #     params = deepcopy(self.meta)
    #     params.update(kwargs)
    #     if which == "edges":
    #         return [ap.fov_grid(which=which, **params) for ap in self.apertures]
    #     if which == "masks":
    #         return {ap.meta["id"]: ap.mask for ap in self.apertures}

    @property
    def apertures(self):
        return self.get_apertures(range(len(self.table)))

    def get_apertures(self, row_ids):
        if isinstance(row_ids, int):
            row_ids = [row_ids]

        apertures_list = []
        for ii in row_ids:
            row = self.table[ii]
            row_dict = {col: row[col] for col in row.colnames}
            row_dict["n_round"] = self.meta["n_round_corners"]
            array_dict = make_aperture_polygon(**row_dict)
            params = {"id": row["id"],
                      "angle": row["angle"],
                      "shape": row["shape"],
                      "conserve_image": yaml.full_load(str(row["conserve_image"])),
                      "no_mask": self.meta["no_mask"],
                      "pixel_scale": self.meta["pixel_scale"],
                      "x_unit": "arcsec",
                      "y_unit": "arcsec",
                      "angle_unit": "arcsec"}
            apertures_list += [ApertureMask(array_dict=array_dict, **params)]

        return apertures_list

    def plot(self):
        import matplotlib.pyplot as plt
        plt.gcf().clf()

        for ap in self.apertures:
            ap.plot(new_figure=False)

        return plt.gcf()

    def plot_masks(self):
        import matplotlib.pyplot as plt

        aps = self.apertures
        n = len(aps)
        w = np.ceil(n ** 0.5).astype(int)
        h = np.ceil(n / w).astype(int)

        for ii, ap in enumerate(aps):
            plt.subplot(w, h, ii + 1)
            plt.imshow(ap.mask.T)
        plt.show()

    def __add__(self, other):
        if isinstance(other, ApertureList):
            from astropy.table import vstack
            self.table = vstack([self.table, other.table])

            return self
        else:
            raise ValueError("Secondary argument not of type ApertureList: {}"
                             "".format(type(other)))

    # def __getitem__(self, item):
    #     return self.get_apertures(item)[0]



class SlitWheel(Effect):
    """
    A selection of predefined spectroscopic slits and possibly other field masks

    It should contain an open position.
    A user can define a non-standard slit by directly using the Aperture
    effect.

    .. todo: This is based on FilterWheel. There is a more efficient way to do this, when we have time.

    Parameters
    ----------
    slit_names : list of str

    filename_format : str
        A f-string for the path to the slit files

    current_slit : str
        Default name

    Examples
    --------
    This Effect assumes a folder full of ASCII files containing the edges of
    each slit. Each file should be names the same except for the slit's name
    or identifier.

    This example assumes a folder ``masks`` containing the slit ASCII files
    with the naming convention: ``slit_A.dat``, ``slit_B.dat``, etc.
    ::

        name: slit_wheel
        class: SlitWheel
        kwargs:
            slit_names:
                - A
                - B
                - C
            filename_format: "masks/slit_{}.dat
            current_slit: "C"

    """
    def __init__(self, **kwargs):
        required_keys = ["slit_names", "filename_format", "current_slit"]
        check_keys(kwargs, required_keys, action="error")

        super(SlitWheel, self).__init__(**kwargs)

        params = {"z_order": [80, 280, 580],
                  "path": "",
                  "report_plot_include": False,
                  "report_table_include": True,
                  "report_table_rounding": 4}
        self.meta.update(params)
        self.meta.update(kwargs)

        path = pth.join(self.meta["path"],
                        from_currsys(self.meta["filename_format"]))
        self.slits = {}
        for name in from_currsys(self.meta["slit_names"]):
            kwargs["name"] = name
            self.slits[name] = ApertureMask(filename=path.format(name),
                                            **kwargs)

        self.table = self.get_table()


    def apply_to(self, obj, **kwargs):
        """Use apply_to of current_slit"""
        return self.current_slit.apply_to(obj, **kwargs)


    def fov_grid(self, which="edges", **kwargs):
        return self.current_slit.fov_grid(which=which, **kwargs)

    def change_slit(self, slitname=None):
        """Change the current slit"""
        if not slitname or slitname in self.slits.keys():
            self.meta['current_slit'] = slitname
            self.include = slitname
        else:
            raise ValueError("Unknown slit requested: " + slitname)

    @property
    def current_slit(self):
        """Return the currently used slit"""
        currslit = from_currsys(self.meta["current_slit"])
        if not currslit:
            return False
        return self.slits[currslit]

    @property
    def display_name(self):
        return f'{self.meta["name"]} : ' \
               f'[{from_currsys(self.meta["current_slit"])}]'


    def __getattr__(self, item):
        return getattr(self.current_slit, item)

    def get_table(self):
        """
        Create a table of slits with centre position, width and length

        Width is defined as the extension in the y-direction, length in the
        x-direction. All values are in milliarcsec.
        """
        names = list(self.slits.keys())
        slits = self.slits.values()
        xmax = np.array([slit.data['x'].max() * u.Unit(slit.meta['x_unit'])
                         .to(u.mas) for slit in slits])
        xmin = np.array([slit.data['x'].min() * u.Unit(slit.meta['x_unit'])
                         .to(u.mas) for slit in slits])
        ymax = np.array([slit.data['y'].max() * u.Unit(slit.meta['y_unit'])
                         .to(u.mas) for slit in slits])
        ymin = np.array([slit.data['y'].min() * u.Unit(slit.meta['y_unit'])
                         .to(u.mas) for slit in slits])
        xmax = quantify(xmax, u.mas)
        xmin = quantify(xmin, u.mas)
        ymax = quantify(ymax, u.mas)
        ymin = quantify(ymin, u.mas)

        lengths = xmax - xmin
        widths = ymax - ymin
        x_centres = (xmax + xmin) / 2
        y_centres = (ymax + ymin) / 2
        tbl = Table(names=["name", "x_centre", "y_centre", "length", "width"],
                    data=[names, x_centres, y_centres, lengths, widths])
        return tbl


################################################################################


def make_aperture_polygon(left, right, top, bottom, angle, shape, **kwargs):

    n_round = kwargs["n_round"] if "n_round" in kwargs else 32
    offset = kwargs["offset"] if "offset" in kwargs else 0.

    n_corners = {"rect": 4, "hex": 6, "oct": 8, "round": n_round}
    try:
        shape = int(float(shape))
        n_corners[shape] = shape
    except:
        pass

    x0, y0 = 0.5 * (right + left), 0.5 * (top + bottom)
    dx, dy = 0.5 * (right - left), 0.5 * (top - bottom)
    n = n_corners[shape]

    if isinstance(shape, str) and "rect" in shape:
        dx *= 1.41421356
        dy *= 1.41421356
        offset += 45.

    x, y = points_on_a_circle(n=n, x0=x0, y0=y0, dx=dx, dy=dy, offset=offset)
    if angle != 0.:
        x, y = rotate(x=x, y=y, x0=np.average(x), y0=np.average(y), angle=angle)

    return {"x": x, "y": y}


def points_on_a_circle(n, x0=0, y0=0, dx=1, dy=1, offset=0):
    deg2rad = np.pi / 180
    d_angle = np.arange(0, 360, 360 / n) + offset
    x = x0 + dx * np.cos(d_angle * deg2rad)
    y = y0 + dy * np.sin(d_angle * deg2rad)

    return x, y


def mask_from_coords(x, y, pixel_scale):
    naxis1 = int(np.ceil((np.max(x) - np.min(x)) / pixel_scale))
    naxis2 = int(np.ceil((np.max(y) - np.min(y)) / pixel_scale))
    xrange = np.linspace(np.min(x), np.max(x), naxis1)
    yrange = np.linspace(np.min(y), np.max(y), naxis2)
    coords = [(xi, yi) for xi in xrange for yi in yrange]

    corners = [(xi, yi) for xi, yi in zip(x, y)]
    path = Path(corners)
    # ..todo: known issue - for super thin apertures, the first row is masked
    # rad = 0.005
    rad = 0  # increase this to include slightly more points within the polygon
    mask = path.contains_points(coords, radius=rad).reshape((naxis2, naxis1))

    return mask


def rotate(x, y, x0, y0, angle):
    """ Rotate a line by ``angle`` [deg] around the point (x0, y0) """
    angle_rad = angle / 57.29578
    xnew = x0 + (x - x0) * np.cos(angle_rad) - (y - y0) * np.sin(angle_rad)
    ynew = y0 + (x - x0) * np.sin(angle_rad) + (y - y0) * np.cos(angle_rad)

    return xnew, ynew



import os
import logging
import copy

import numpy as np
import yaml
import requests

from .. import rc
from ..utils import find_file

__all__ = ["UserCommands"]


class UserCommands:
    """
    Contains all the setting that a user may wish to alter for an optical train

    Most of the important settings are kept in the ``.cmds`` system dictionary
    Setting can be accessed by using the alias names. Currently these are:

    - ATMO: atmospheric and observatory location settings
    - TEL: telescope related settings
    - RO: relay optics settings, i.e. between telescope and instrument
    - INST: instrument optics settings
    - DET: detector settings
    - OBS: observation settings, and
    - SIM: simulation settings

    All of the settings are contained in a special ``SystemDict`` dictionary
    that allows the user to access all the settings via a bang-string (!). E.g::

        cmds = UserCommands()
        cmds["!SIM.file.local_packages_path]

    .. note::

       To use this format for accessing hierarchically-stored values, the bang
       string must always begin with a "!"

    Alternatively the same value can be accessed via the normal dictionary
    format. E.g::

        cmds["SIM"]["file"]["local_packages_path"]


    Parameters
    ----------
    use_instrument : str, optional
        The name of the main instrument to use

    packages : list, optional
        list of package names needed for the optical system, so that ScopeSim
        can find the relevant files. E.g. ["Armazones", "ELT", "MICADO"]

    yamls : list, optional
        list of yaml filenames that are needed for the combined optical system
        E.g. ["MICADO_Standalone_RO.yaml", "MICADO_H4RG.yaml", "MICADO_.yaml"]

    mode_yamls : list of yamls, optional
        list of yaml docs ("OBS" docs) that are applicable only to specific
        operational modes of the instrument.
        Further yaml files can be specified in the recursive doc entry: "yamls"

    set_modes : list of strings, optional
        A list of default mode yamls to load. E.g. ["SCAO", "IMG_4mas"]

    properties : dict, optional
        Any extra "OBS" properties that should be added

    ignore_effects : list
        Not yet implemented

    add_effects : list
        Not yet implemented

    override_effect_values : dict
        Not yet implemented


    Attributes
    ----------
    cmds : SystemDict
        Built from the ``properties`` dictionary of a yaml dictionary. All
        values here are accessible globally by all ``Effects`` objects in an
        ``OpticalTrain`` once the ``UserCommands`` has been passed to the
        ``OpticalTrain``.

    yaml_dicts : list of dicts
        Where all the effects dictionaries are stored


    Examples
    --------
    Here we use a combination of the main parameters: ``packages``, ``yamls``,
    and ``properties``. When not using the ``use_instrument`` key, ``packages``
    and ``yamls`` must be specified, otherwise scopesim will not know
    where to look for yaml files (only relevant if reading in yaml files)::

        >>> from scopesim.server.database import download_package
        >>> from scopesim.commands import UserCommands
        >>>
        >>> download_package("test_package")
        >>> cmd = UserCommands(packages=["test_package"],
        ...                    yamls=["test_telescope.yaml",
        ...                           {"alias": "ATMO",
        ...                            "properties": {"pwv": 9001}}],
        ...                    properties={"!ATMO.pwv": 8999})

    Notes
    -----
    .. attention:: We track your IP address when ``ScopeSim`` checks for updates

        When initialising a UserCommands object via ``use_instrument=``,
        ``ScopeSim`` checks on the database whether there are updates to the
        instrument package. Our server records the IP address of each query for
        out own statistics only.

        WE DO NOT STORE OR TRACK PERSONAL DATA. THESE STATISTICS ARE NEEDED FOR
        GETTING MORE FUNDING TO CONTINUE DEVELOPING THIS PROJECT.

        We are doing this solely as a way of showing the austrian funding agency
        that people are indeed using this software (or not). Your participation
        in this effort greatly helps our chances of securing the next grant.

        However, if you would still like to avoid your IP address being stored,
        you can run ``scopesim`` 100% anonymously by setting::

            >>> scopsim.rc.__config__["!SIM.reports.ip_tracking"] = True

        at the beginning of each session. Alternatively you can also pass the
        same bang keyword when generating a ``UserCommand`` object::

            >>> from scopesim import UserCommands
            >>> UserCommands(use_instrument="MICADO",
            >>>              properties={"!SIM.reports.ip_tracking": False})

        If you use a custom ``yaml`` configuration file, you can also add this
        keyword to the ``properties`` section of the ``yaml`` file.

    """

    def __init__(self, **kwargs):

        self.cmds = copy.deepcopy(rc.__config__)
        self.yaml_dicts = []
        self.kwargs = kwargs
        self.ignore_effects = []
        self.package_name = ""
        self.default_yamls = []
        self.modes_dict = {}

        self.update(**kwargs)

    def update(self, **kwargs):
        """
        Updates the current parameters with a yaml dictionary

        See the ``UserCommands`` main docstring for acceptable kwargs
        """

        if "use_instrument" in kwargs:
            self.package_name = kwargs["use_instrument"]
            self.update(packages=[kwargs["use_instrument"]],
                        yamls=["default.yaml"])

            check_for_updates(self.package_name)

        if "packages" in kwargs:
            add_packages_to_rc_search(self["!SIM.file.local_packages_path"],
                                      kwargs["packages"])

        if "yamls" in kwargs:
            for yaml_input in kwargs["yamls"]:
                if isinstance(yaml_input, str):
                    yaml_file = find_file(yaml_input)
                    if yaml_file is not None:
                        yaml_dict = load_yaml_dicts(yaml_file)
                        self.update(yamls=yaml_dict)
                        if yaml_input == "default.yaml":
                            self.default_yamls = yaml_dict
                    else:
                        logging.warning("{} could not be found".format(yaml_input))

                elif isinstance(yaml_input, dict):
                    self.cmds.update(yaml_input)
                    self.yaml_dicts += [yaml_input]

                    for key in ["packages", "yamls", "mode_yamls"]:
                        if key in yaml_input:
                            self.update(**{key: yaml_input[key]})

                else:
                    raise ValueError("yaml_dicts must be a filename or a "
                                     "dictionary: {}".format(yaml_input))

        if "mode_yamls" in kwargs:
            # Convert the yaml list of modes to a dict object
            self.modes_dict = {my["name"]: my for my in kwargs["mode_yamls"]}
            if "modes" in self.cmds["!OBS"]:
                if not isinstance(self.cmds["!OBS.modes"], list):
                    self.cmds["!OBS.modes"] = [self.cmds["!OBS.modes"]]
                for mode_name in self.cmds["!OBS.modes"]:
                    mode_yaml = self.modes_dict[mode_name]
                    self.update(yamls=[mode_yaml])

        if "set_modes" in kwargs:
            self.set_modes(modes=kwargs["set_modes"])

        if "properties" in kwargs:
            self.cmds.update(kwargs["properties"])

        if "ignore_effects" in kwargs:
            self.ignore_effects = kwargs["ignore_effects"]

        if "add_effects" in kwargs:
            # ..todo: implement this
            pass

        if "override_effect_values" in kwargs:
            # ..todo: implement this
            pass

    def set_modes(self, modes=None):
        if not isinstance(modes, list):
            modes = [modes]
        for defyam in self.default_yamls:
            if "properties" in defyam and "modes" in defyam["properties"]:
                defyam["properties"]["modes"] = []
                for mode in modes:
                    if mode in self.modes_dict:
                        defyam["properties"]["modes"] += [mode]
                    else:
                        raise ValueError("mode '{}' was not recognised"
                                         "".format(mode))

        self.__init__(yamls=self.default_yamls)

    def list_modes(self):
        if isinstance(self.modes_dict, dict):
            modes = {}
            for mode_name in self.modes_dict:
                dic = self.modes_dict[mode_name]
                desc = dic["description"] if "description" in dic else "<None>"
                modes[mode_name] = desc

            msg = "\n".join(["{}: {}".format(key, modes[key]) for key in modes])
        else:
            msg = "No modes found"
        return msg

    @property
    def modes(self):
        print(self.list_modes())

    def __setitem__(self, key, value):
        self.cmds.__setitem__(key, value)

    def __getitem__(self, item):
        return self.cmds.__getitem__(item)

    def __contains__(self, item):
        return self.cmds.__contains__(item)

    def __repr__(self):
        return self.cmds.__repr__()


def check_for_updates(package_name):
    """
    Asks IRDB server if there are newer versions of the instrument package
    """
    response = {}

    # tracking **exclusively** your IP address for our internal stats
    if rc.__currsys__["!SIM.reports.ip_tracking"] and \
            "TRAVIS" not in os.environ:
        front_matter = rc.__currsys__["!SIM.file.server_base_url"]
        back_matter = "api.php?package_name={}".format(package_name)
        try:
            response = requests.get(url=front_matter+back_matter).json()
        except:
            print("Offline. Cannot check for updates for {}"
                  "".format(package_name))
    return response


def add_packages_to_rc_search(local_path, package_list):
    """
    Adds the paths of a list of locally saved packages to the search path list

    Parameters
    ----------
    local_path : str
        Where the pacakges are located. Generally given by the value in
        scopesim.rc.__config__["!SIM.file.local_packages_path"]

    package_list : list
        A list of the package names to add

    """

    for pkg in package_list:
        pkg_dir = os.path.abspath(os.path.join(local_path, pkg))
        if not os.path.exists(pkg_dir):
            # todo: keep here, but add test for this by downloading test_package
            # raise ValueError("Package could not be found: {}".format(pkg_dir))
            logging.warning("Package could not be found: {}".format(pkg_dir))

        if pkg_dir in rc.__search_path__:
            # if package is already in search_path, move it to the first place
            ii = np.where(np.array(rc.__search_path__) == pkg_dir)[0][0]
            rc.__search_path__.pop(ii)

        rc.__search_path__ = [pkg_dir] + rc.__search_path__


def load_yaml_dicts(filename):
    """
    Loads one or more dicts stored in a YAML file under ``filename``

    Parameters
    ----------
    filename : str
        Path to the YAML file

    Returns
    -------
    yaml_dicts : list
        A list of dicts

    """

    yaml_dicts = []
    with open(filename) as f:
        yaml_dicts += [dic for dic in yaml.full_load_all(f)]

    return yaml_dicts


def list_local_packages(action="display"):
    """
    Lists the packages on the local disk that ScopeSim can find

    Packages can only be found in the directory listed under::

        scopesim.rc.__config__["!SIM.file.local_packages_path"]

    Packages are divided into "main" packages and "extension" packages.

    - Main packages contain a ``default.yaml`` file which tell ScopeSim which
      other packages are required to generate the full optical system
    - Extension packages contain only the data files needed to support the
      effects listed in the package YAML file

    .. note::
       Only "main" packages can be passed to a UserCommands object using the
       ``use_instrument=...`` parameter

    Parameters
    ----------
    action : str, optional
        ["display", "return"] What to do with the output.
        - "display": the list of packages are printed to the screen
        - "return": package names are returned in lists

    Returns
    -------
    main_pkgs, ext_pkgs : lists
        If action="return": Lists containing the names of locally saved packages

    """

    local_path = os.path.abspath(rc.__config__["!SIM.file.local_packages_path"])
    pkgs = [d for d in os.listdir(local_path) if
            os.path.isdir(os.path.join(local_path, d))]

    main_pkgs = [pkg for pkg in pkgs if
                 os.path.exists(os.path.join(local_path, pkg, "default.yaml"))]
    ext_pkgs = [pkg for pkg in pkgs if not
                os.path.exists(os.path.join(local_path, pkg, "default.yaml"))]

    if action == "display":
        msg = "\nLocal package directory:\n  {}\n" \
              "Full packages [can be used with 'use_instrument=...']\n  {}\n" \
              "Support packages\n  {}".format(local_path, main_pkgs, ext_pkgs)
        print(msg)
    else:
        return main_pkgs, ext_pkgs



import numpy as np
from astropy import units as u
from os import path as pth
import logging

from astropy.io import fits
from astropy.table import Table
from astropy import units as u

from synphot import SourceSpectrum
from synphot.units import PHOTLAM
import skycalc_ipy

from .ter_curves_utils import combine_two_spectra, apply_throughput_to_cube
from .ter_curves_utils import download_svo_filter, download_svo_filter_list
from .ter_curves_utils import add_edge_zeros
from .effects import Effect
from ..optics.surface import SpectralSurface
from ..source.source_utils import make_imagehdu_from_table
from ..source.source import Source
from ..base_classes import SourceBase, FOVSetupBase
from ..utils import from_currsys, quantify, check_keys, find_file


class TERCurve(Effect):
    """
    Transmission, Emissivity, Reflection Curve

    Must contain a wavelength column, and one or more of the following:
    ``transmission``, ``emissivity``, ``reflection``.
    Additionally in the header there
    should be the following keywords: wavelength_unit

    kwargs that can be passed::

        "rescale_emission" : { "filter_name": str, "value": float, "unit": str}

    Examples
    --------
    Directly inside a YAML file description::

        name: bogus_surface
        class: TERCurve
        kwargs:
            array_dict:
                wavelength: [0.3, 3.0]
                transmission: [0.9, 0.9]
                emission: [1, 1]
            wavelength_unit: um
            emission_unit: ph s-1 m-2 um-1
            rescale_emission:
                filter_name: "Paranal/HAWKI.Ks"
                value: 15.5
                unit: ABmag

    Indirectly inside a YAML file::

        name: some_curve
        class TERCurve
        kwargs:
            filename: bogus_surface.dat

    which references this ASCII file::

        # name: bogus_surface
        # wavelength_unit: um
        wavelength  transmission    emissivity
        0.3         0.9             0.1
        3.0         0.9             0.1

    """
    def __init__(self, **kwargs):
        super(TERCurve, self).__init__(**kwargs)
        params = {"z_order": [10, 110, 510],
                  "ignore_wings": False,
                  "wave_min": "!SIM.spectral.wave_min",
                  "wave_max": "!SIM.spectral.wave_max",
                  "wave_unit": "!SIM.spectral.wave_unit",
                  "wave_bin": "!SIM.spectral.spectral_bin_width",
                  "bg_cell_width": "!SIM.computing.bg_cell_width",
                  "report_plot_include": True,
                  "report_table_include": False}
        self.meta.update(params)
        self.meta.update(kwargs)

        self.surface = SpectralSurface()
        self.surface.meta.update(self.meta)
        self._background_source = None

        data = self.get_data()
        if self.meta["ignore_wings"]:
            data = add_edge_zeros(data, "wavelength")
        if data is not None:
            self.surface.table = data
            self.surface.table.meta.update(self.meta)

    # ####### added in new branch

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, SourceBase):
            self.meta = from_currsys(self.meta)
            wave_min = quantify(self.meta["wave_min"], u.um).to(u.AA)
            wave_max = quantify(self.meta["wave_max"], u.um).to(u.AA)

            thru = self.throughput

            # apply transmission to source spectra
            for isp, spec in enumerate(obj.spectra):
                obj.spectra[isp] = combine_two_spectra(spec, thru, "multiply",
                                                       wave_min, wave_max)

            # apply transmission to cube fields
            for icube, cube in enumerate(obj.cube_fields):
                obj.cube_fields[icube] = apply_throughput_to_cube(cube, thru)

            # add the effect background to the source background field
            if self.background_source is not None:
                obj.append(self.background_source)

        if isinstance(obj, FOVSetupBase):
            wave = self.surface.throughput.waveset
            thru = self.surface.throughput(wave)
            valid_waves = np.argwhere(thru > 0)
            wave_min = wave[max(0, valid_waves[0][0] - 1)]
            wave_max = wave[min(len(wave) - 1, valid_waves[-1][0] + 1)]

            obj.shrink("wave", [wave_min.to(u.um).value,
                                wave_max.to(u.um).value])

        return obj

    @property
    def emission(self):
        return self.surface.emission

    @property
    def throughput(self):
        return self.surface.throughput

    @property
    def background_source(self):
        if self._background_source is None:
            # add a single pixel ImageHDU for the extended background with a
            # size of 1 degree
            bg_cell_width = from_currsys(self.meta["bg_cell_width"])
            flux = self.emission
            bg_hdu = fits.ImageHDU()
            bg_hdu.header.update({"BG_SRC": True,
                                  "BG_SURF": self.display_name,
                                  "CUNIT1": "ARCSEC",
                                  "CUNIT2": "ARCSEC",
                                  "CDELT1": 0,
                                  "CDELT2": 0,
                                  "BUNIT": "PHOTLAM arcsec-2",
                                  "SOLIDANG": "arcsec-2"})
            self._background_source = Source(image_hdu=bg_hdu, spectra=flux)

        return self._background_source

    # #######

    def plot(self, which="x", wavelength=None, ax=None, new_figure=True,
             label=None, **kwargs):
        """

        Parameters
        ----------
        which : str
            "x" plots throughput. "t","e","r" plot trans/emission/refl
        wavelength : list, np.ndarray
        ax : matplotlib.Axis
        kwargs

        Returns
        -------

        """
        import matplotlib.pyplot as plt
        if new_figure:
            # from matplotlib import rcParams
            # rcParams["axes.formatter.useoffset"] = False
            plt.figure(figsize=(10, 5))

        self.meta.update(kwargs)
        params = from_currsys(self.meta)

        for ii, ter in enumerate(which):
            if ax is None:
                plt.subplot(len(which), 1, ii+1)

            if wavelength is None:
                wunit = params["wave_unit"]
                wave = np.arange(quantify(params["wave_min"], wunit).value,
                                 quantify(params["wave_max"], wunit).value,
                                 quantify(params["wave_bin"], wunit).value)
                wave *= u.Unit(wunit)
            else:
                wave = wavelength

            plot_kwargs = self.meta.get("plot_kwargs", {})
            surf = self.surface

            if "t" in ter:
                y = surf.transmission(wave)
            elif "e" in ter:
                y = surf.emission(wave)
            elif "r" in ter:
                y = surf.reflection(wave)
            else:
                y = surf.throughput(wave)

            plt.plot(wave, y, **plot_kwargs)

            wave_unit = self.meta.get("wavelength_unit")
            plt.xlabel("Wavelength [{}]".format(wave_unit))
            y_str = {"t": "Transmission", "e": "Emission",
                     "r": "Reflectivity", "x": "Throughput"}
            plt.ylabel("{} [{}]".format(y_str[ter], y.unit))

        return plt.gcf()


class AtmosphericTERCurve(TERCurve):
    def __init__(self, **kwargs):
        super(AtmosphericTERCurve, self).__init__(**kwargs)
        self.meta["z_order"] = [111, 511]
        self.meta["action"] = "transmission"
        self.meta["position"] = 0       # position in surface table
        self.meta.update(kwargs)
        self.surface.meta.update(self.meta)


class SkycalcTERCurve(AtmosphericTERCurve):
    def __init__(self, **kwargs):
        """
        Retrieves an atmospheric spectrum from ESO's skycalc server

        kwarg parameters
        ----------------
        skycalc parameters can be found by calling::

            >>> import skycalc_ipy
            >>> skycalc_ipy.SkyCalc().keys

        .. note:: Different to skycalc_ipy, wmin and wmax must be given in units
            of ``um``

        Examples
        --------
        ::

            - name : skycalc_background
              class : SkycalcTERCurve
              kwargs :
                wunit : "!SIM.spectral.wave_unit"
                wmin : "!SIM.spectral.wave_min"
                wmax : "!SIM.spectral.wave_max"
                wdelta : 0.0001     # 0.1nm bin width
                outer : 1
                outer_unit : "m"

        """
        super(SkycalcTERCurve, self).__init__(**kwargs)
        self.meta["z_order"] = [112, 512]
        self.meta["use_local_skycalc_file"] = False
        self.meta.update(kwargs)

        self.skycalc_table = None

        if self.include is True:
            self.load_skycalc_table()

    @property
    def include(self):
        return from_currsys(self.meta["include"])

    @include.setter
    def include(self, item):
        self.meta["include"] = item
        if item is True and self.skycalc_table is None:
            self.load_skycalc_table()

    def load_skycalc_table(self):
        use_local_file = from_currsys(self.meta["use_local_skycalc_file"])
        if not use_local_file:
            self.skycalc_conn = skycalc_ipy.SkyCalc()
            tbl = self.query_server()

            if "name" not in self.meta:
                self.meta["name"] = self.skycalc_conn["observatory"]

        else:
            path = find_file(use_local_file)
            fits_tbl = fits.getdata(path, ext=1)
            fits_hdr = fits.getheader(path, ext=0)
            tbl = Table(fits_tbl)
            tbl["lam"].unit = u.um
            for colname in tbl.colnames:
                if "flux" in colname:
                    tbl[colname].unit = u.Unit("ph s-1 m-2 um-1 arcsec-2")
            tbl_small = Table()
            tbl_small.meta["fits_header"] = dict(fits_hdr)
            tbl_small.add_columns([tbl["lam"], tbl["trans"], tbl["flux"]])
            tbl = tbl_small

        for i, colname in enumerate(["wavelength", "transmission", "emission"]):
            tbl.columns[i].name = colname
        tbl.meta["wavelength_unit"] = tbl.columns[0].unit
        tbl.meta["emission_unit"] = tbl.columns[2].unit

        self.surface.table = tbl
        self.surface.meta.update(tbl.meta)
        self.skycalc_table = tbl

    def query_server(self, **kwargs):
        self.meta.update(kwargs)

        if "wunit" in self.meta:
            scale_factor = u.Unit(from_currsys(self.meta["wunit"])).to(u.nm)
            for key in ["wmin", "wmax", "wdelta"]:
                if key in self.meta:
                    self.meta[key] = from_currsys(self.meta[key]) * scale_factor

        conn_kwargs = {key: self.meta[key] for key in self.meta
                       if key in self.skycalc_conn.defaults}
        conn_kwargs = from_currsys(conn_kwargs)
        self.skycalc_conn.values.update(conn_kwargs)

        try:
            tbl = self.skycalc_conn.get_sky_spectrum(return_type="table")
        except:
            msg = "Could not connect to skycalc server"
            logging.exception(msg)
            raise ValueError(msg)

        return tbl


class QuantumEfficiencyCurve(TERCurve):
    def __init__(self, **kwargs):
        super(QuantumEfficiencyCurve, self).__init__(**kwargs)
        self.meta["action"] = "transmission"
        self.meta["z_order"] = [113, 513]
        self.meta["position"] = -1          # position in surface table


class FilterCurve(TERCurve):
    """
    Parameters
    ----------
    position : int, optional
    filter_name : str, optional
        ``Ks`` - corresponding to the filter name in the filename pattern
    filename_format : str, optional
        ``TC_filter_{}.dat``

    Can either be created using the standard 3 options:
    - ``filename``: direct filename of the filer curve
    - ``table``: an ``astropy.Table``
    - ``array_dict``: a dictionary version of a table: ``{col_name1: values, }``

    or by passing the combination of ``filter_name`` and ``filename_format`` as
    kwargs. Here all filter file names follow a pattern (e.g. see above) and the
    ``{}`` are replaced by ``filter_name`` at run time. ``filter_name`` can
    also be a !bang string for a ``__currsys__`` entry: ``"!INST.filter_name"``

    """
    def __init__(self, **kwargs):
        if not np.any([key in kwargs for key in ["filename", "table",
                                                 "array_dict"]]):
            if "filter_name" in kwargs and "filename_format" in kwargs:
                filt_name = from_currsys(kwargs["filter_name"])
                file_format = from_currsys(kwargs["filename_format"])
                kwargs["filename"] = file_format.format(filt_name)
            else:
                raise ValueError("FilterCurve must be passed one of (`filename`"
                                 " `array_dict`, `table`) or both "
                                 "(`filter_name`, `filename_format`):"
                                 "{}".format(kwargs))

        super(FilterCurve, self).__init__(**kwargs)
        if self.table is None:
            raise ValueError("Could not initialise filter. Either filename not "
                             "found, or array are not compatible")

        params = {"minimum_throughput": "!SIM.spectral.minimum_throughput",
                  "action": "transmission",
                  "position": -1,               # position in surface table
                  "wing_flux_level": None,
                  "name": "untitled filter"}
        self.meta.update(params)
        self.meta["z_order"] = [114, 214, 514]
        self.meta.update(kwargs)

        min_thru = from_currsys(self.meta["minimum_throughput"])
        mask = self.table["transmission"] < min_thru
        self.table["transmission"][mask] = 0

    def fov_grid(self, which="waveset", **kwargs):
        if which == "waveset":
            self.meta.update(kwargs)
            self.meta = from_currsys(self.meta)
            # ..todo:: replace the 101 with a variable in !SIM
            wave = np.linspace(self.meta["wave_min"],
                               self.meta["wave_max"], 101)
            wave = quantify(wave, u.um)
            throughput = self.surface.transmission(wave)
            min_thru = self.meta["minimum_throughput"]
            valid_waves = np.where(throughput.value > min_thru)[0]
            if len(valid_waves) > 0:
                wave_edges = [min(wave[valid_waves].value),
                              max(wave[valid_waves].value)] * u.um
            else:
                raise ValueError("No transmission found above the threshold {} "
                                 "in this wavelength range {}. Did you open "
                                 "the shutter?"
                                 "".format(self.meta["minimum_throughput"],
                                           [self.meta["wave_min"],
                                            self.meta["wave_max"]]))
        else:
            wave_edges = []

        return wave_edges

    @property
    def fwhm(self):
        wave = self.surface.wavelength
        thru = self.surface._get_ter_property("transmission", fmt="array")
        mask = thru >= 0.5
        if any(mask):
            dwave = wave[mask][-1] - wave[mask][0]
        else:
            dwave = 0 * wave.unit

        return dwave

    @property
    def centre(self):
        wave = self.surface.wavelength
        thru = self.surface._get_ter_property("transmission", fmt="array")
        num = np.trapz(thru * wave**2, x=wave)
        den = np.trapz(thru * wave, x=wave)

        return num / den

    @property
    def center(self):
        return self.centre


class TopHatFilterCurve(FilterCurve):
    """
    A simple Top-Hat filter profile

    Parameters
    ----------
    transmission : float
        [0..1] Peak transmission of filter

    blue_cutoff, red_cutoff : float
        [um] Blue and Red cutoff wavelengths

    wing_transmission : float, optional
        [0..1] Default 0. Wing transmission of filter outside the cutoff range

    Examples
    --------
    ::

        name: J_band_tophat
        class: TopHatFilterCurve
        kwargs:
            transmission : 0.9
            wing_transmission : 0.001
            blue_cutoff : 1.15
            red_cutoff : 1.35


    """
    def __init__(self, **kwargs):
        required_keys = ["transmission", "blue_cutoff", "red_cutoff"]
        check_keys(kwargs, required_keys, action="error")

        wave_min = from_currsys("!SIM.spectral.wave_min")
        wave_max = from_currsys("!SIM.spectral.wave_max")
        blue = kwargs["blue_cutoff"]
        red = kwargs["red_cutoff"]
        peak = kwargs["transmission"]
        wing = kwargs.get("wing_transmission", 0)
        
        waveset = [wave_min, 0.999*blue, blue, red, red*1.001, wave_max]
        transmission = [wing, wing, peak, peak, wing, wing]
        
        tbl = Table(names=["wavelength", "transmission"],
                    data=[waveset, transmission])
        super(TopHatFilterCurve, self).__init__(table=tbl,
                                                wavelength_unit="um",
                                                action="transmission")
        self.meta.update(kwargs)


class DownloadableFilterCurve(FilterCurve):
    def __init__(self, **kwargs):
        required_keys = ["filter_name", "filename_format"]
        check_keys(kwargs, required_keys, action="error")
        filt_str = kwargs["filename_format"].format(kwargs["filter_name"])
        tbl = download_svo_filter(filt_str, return_style="table")
        super(DownloadableFilterCurve, self).__init__(table=tbl, **kwargs)


class SpanishVOFilterCurve(FilterCurve):
    """
    Pulls a filter transmission curve down from the Spanish VO filter service

    Parameters
    ----------
    observatory : str
    instrument : str
    filter_name : str

    Examples
    --------
    ::

        name: HAWKI-Ks
        class: SpanishVOFilterCurve
        kwargs:
            observatory : Paranal
            instrument : HAWKI
            filter_name : Ks

    """
    def __init__(self, **kwargs):
        required_keys = ["observatory", "instrument", "filter_name"]
        check_keys(kwargs, required_keys, action="error")
        filt_str = "{}/{}.{}".format(kwargs["observatory"],
                                     kwargs["instrument"],
                                     kwargs["filter_name"])
        kwargs["name"] = kwargs["filter_name"]
        kwargs["svo_id"] = filt_str

        raise_error = kwargs.get("error_on_wrong_name", True)
        tbl = download_svo_filter(filt_str, return_style="table",
                                  error_on_wrong_name=raise_error)
        super(SpanishVOFilterCurve, self).__init__(table=tbl, **kwargs)


class FilterWheel(Effect):
    """
    This wheel holds a selection of predefined filters.

    Examples
    --------
    ::

        name: filter_wheel
        class: FilterWheel
        kwargs:
            filter_names: []
            filename_format: "filters/{}.
            current_filter: "Ks"

    """

    def __init__(self, **kwargs):
        required_keys = ["filter_names", "filename_format", "current_filter"]
        check_keys(kwargs, required_keys, action="error")

        super(FilterWheel, self).__init__(**kwargs)

        params = {"z_order": [124, 224, 524],
                  "path": "",
                  "report_plot_include": True,
                  "report_table_include": True,
                  "report_table_rounding": 4}
        self.meta.update(params)
        self.meta.update(kwargs)

        path = pth.join(self.meta["path"],
                        from_currsys(self.meta["filename_format"]))
        self.filters = {}
        for name in from_currsys(self.meta["filter_names"]):
            kwargs["name"] = name
            self.filters[name] = FilterCurve(filename=path.format(name),
                                             **kwargs)

        self.table = self.get_table()


    def apply_to(self, obj, **kwargs):
        """Use apply_to of current filter"""
        return self.current_filter.apply_to(obj, **kwargs)

    def fov_grid(self, which="waveset", **kwargs):
        return self.current_filter.fov_grid(which=which, **kwargs)

    def change_filter(self, filtername=None):
        """Change the current filter"""
        if filtername in self.filters.keys():
            self.meta['current_filter'] = filtername
        else:
            raise ValueError("Unknown filter requested: " + filtername)

    @property
    def current_filter(self):
        filter_eff = None
        filt_name = from_currsys(self.meta["current_filter"])
        if filt_name is not None:
            filter_eff = self.filters[filt_name]
        return filter_eff

    @property
    def display_name(self):
        return f'{self.meta["name"]} : ' \
               f'[{from_currsys(self.meta["current_filter"])}]'

    def __getattr__(self, item):
        return getattr(self.current_filter, item)

    def plot(self, which="x", wavelength=None, **kwargs):
        """

        Parameters
        ----------
        which : str
            "x" plots throughput. "t","e","r" plot trans/emission/refl
        wavelength
        kwargs

        Returns
        -------

        """
        import matplotlib.pyplot as plt
        plt.figure(figsize=(10, 5))

        for ii, ter in enumerate(which):
            ax = plt.subplot(len(which), 1, ii+1)
            for name in self.filters:
                self.filters[name].plot(which=ter, wavelength=wavelength,
                                        ax=ax, new_figure=False,
                                        plot_kwargs={"label": name}, **kwargs)

        # plt.semilogy()
        plt.legend()

        return plt.gcf()

    def get_table(self):
        names = list(self.filters.keys())
        ters = self.filters.values()
        centres = u.Quantity([ter.centre for ter in ters])
        widths = u.Quantity([ter.fwhm for ter in ters])
        blue = centres - 0.5 * widths
        red = centres + 0.5 * widths

        tbl = Table(names=["name", "centre", "width", "blue cutoff", "red cutoff"],
                    data=[names, centres, widths, blue, red])

        return tbl


class TopHatFilterWheel(FilterWheel):
    """
    A selection of top-hat filter curves as defined in the input lists

    Parameters
    ----------
    filter_names: list of string

    transmissions: list of floats
        [0..1] Peak transmissions inside the cuttoff limits

    wing_transmissions: list of floats
        [0..1] Wing transmissions outside the cuttoff limits

    blue_cutoffs: list of floats
        [um]

    red_cutoffs: list of floats
        [um]

    current_filter: str, optional
        Name of current filter at initialisation. If no name is given, the
        first entry in ``filter_names`` is used by default.

    Examples
    --------
    ::

        name: top_hat_filter_wheel
        class: TopHatFilterWheel
        kwargs:
            filter_names: ["J", "H", "K"]
            transmissions: [0.9, 0.95, 0.85]
            wing_transmissions: [0., 0., 0.001]
            blue_cutoffs: [1.15, 1.45, 1.9]
            red_cutoffs: [1.35, 1.8, 2.4]
            current_filter: "K"

    """
    def __init__(self, **kwargs):
        required_keys = ["filter_names", "transmissions", "wing_transmissions",
                         "blue_cutoffs", "red_cutoffs"]
        check_keys(kwargs, required_keys, action="error")

        super(FilterWheel, self).__init__(**kwargs)

        params = {"z_order": [124, 224, 524],
                  "report_plot_include": True,
                  "report_table_include": True,
                  "report_table_rounding": 4,
                  "current_filter": kwargs.get("current_filter",
                                               kwargs["filter_names"][0])
                  }
        self.meta.update(params)
        self.meta.update(kwargs)

        self.filters = {}
        for i in range(len(self.meta["filter_names"])):
            name = self.meta["filter_names"][i]
            effect_kwargs = {"name": name,
                             "transmission": self.meta["transmissions"][i],
                             "wing_transmission": self.meta["wing_transmissions"][i],
                             "blue_cutoff": self.meta["blue_cutoffs"][i],
                             "red_cutoff": self.meta["red_cutoffs"][i]}
            self.filters[name] = TopHatFilterCurve(**effect_kwargs)


class SpanishVOFilterWheel(FilterWheel):
    """
    A FilterWheel that loads all the filters from the Spanish VO service

    .. warning::
       This use ``astropy.download_file(..., cache=True)``.

    The filter transmission curves probably won't change, but if you notice
    discrepancies, try clearing the astopy cache::

        >> from astropy.utils.data import clear_download_cache
        >> clear_download_cache()

    Parameters
    ----------
    observatory : str

    instrument : str

    current_filter : str
        Default filter name

    include_str, exclude_str : str
        String sequences that can be used to include or exclude filter names
        which contain a certain string.
        E.g. GTC/OSIRIS has curves for ``sdss_g`` and ``sdss_g_filter``.
        We can force the inclusion of only the filter curves by setting
        ``list_include_str: "_filter"``.

    Examples
    --------
    ::

        name: svo_filter_wheel
        class: SpanishVOFilterWheel
        kwargs:
            observatory: "GTC"
            instrument: "OSIRIS"
            current_filter: "sdss_r_filter"
            include_str: "_filter"

    """
    def __init__(self, **kwargs):
        required_keys = ["observatory", "instrument", "current_filter"]
        check_keys(kwargs, required_keys, action="error")

        # Call Effect.init, *NOT* FilterWheel.init --> different required_keys
        super(FilterWheel, self).__init__(**kwargs)

        params = {"z_order": [124, 224, 524],
                  "report_plot_include": True,
                  "report_table_include": True,
                  "report_table_rounding": 4,
                  "include_str": None,         # passed to
                  "exclude_str": None,
                  }
        self.meta.update(params)
        self.meta.update(kwargs)

        obs, inst = self.meta["observatory"], self.meta["instrument"]
        inc, exc =  self.meta["include_str"], self.meta["exclude_str"]
        filter_names = download_svo_filter_list(obs, inst, short_names=True,
                                                include=inc, exclude=exc)

        self.meta["filter_names"] = filter_names
        self.filters = {name: SpanishVOFilterCurve(observatory=obs,
                                                   instrument=inst,
                                                   filter_name=name)
                        for name in filter_names}
        self.filters["open"] = FilterCurve(array_dict={"wavelength": [0.3, 3.0],
                                                       "transmission": [1., 1.]},
                                           wavelength_unit="um",
                                           name="unity transmission")

        self.table = self.get_table()


class PupilTransmission(TERCurve):
    """
    Wavelength-independent transmission curve

    Use this class to describe a cold stop or pupil mask that is
    characterised by "grey" transmissivity.
    The emissivity is set to zero, assuming that the mask is cold.
    """
    def __init__(self, transmission, **kwargs):
        self.params = {"wave_min": "!SIM.spectral.wave_min",
                       "wave_max": "!SIM.spectral.wave_max"}
        self.params.update(kwargs)
        self.make_ter_curve(transmission)

    def update_transmission(self, transmission, **kwargs):
        self.params.update(kwargs)
        self.make_ter_curve(transmission)

    def make_ter_curve(self, transmission):
        wave_min = from_currsys(self.params["wave_min"]) * u.um
        wave_max = from_currsys(self.params["wave_max"]) * u.um
        transmission = from_currsys(transmission)

        super().__init__(wavelength=[wave_min, wave_max],
                         transmission=[transmission, transmission],
                         emissivity=[0., 0.], **self.params)


class ADCWheel(Effect):
    """
    This wheel holds a selection of predefined atmospheric dispersion
    correctors.

    Example
    -------
    ::

       name : adc_wheel
       class: ADCWheel
       kwargs:
           adc_names: []
           filename_format: "TER_ADC_{}.dat"
           current_adc: "const_90"
    """
    def __init__(self, **kwargs):
        required_keys = ["adc_names", "filename_format", "current_adc"]
        check_keys(kwargs, required_keys, action="error")

        super().__init__(**kwargs)

        params = {"z_order": [125, 225, 525],
                  "path": "",
                  "report_plot_include": False,
                  "report_table_include": True,
                  "report_table_rounding": 4}
        self.meta.update(params)
        self.meta.update(kwargs)

        path = pth.join(self.meta["path"],
                        from_currsys(self.meta["filename_format"]))
        self.adcs = {}
        for name in from_currsys(self.meta["adc_names"]):
            kwargs["name"] = name
            self.adcs[name] = TERCurve(filename=path.format(name),
                                       **kwargs)

        self.table = self.get_table()

    def apply_to(self, obj, **kwargs):
        """Use apply_to of current adc"""
        return self.current_adc.apply_to(obj, **kwargs)

    def change_adc(self, adcname=None):
        """Change the current ADC"""
        if not adcname or adcname in self.adcs.keys():
            self.meta['current_adc'] = adcname
            self.include = adcname
        else:
            raise ValueError("Unknown ADC requested: " + adcname)

    @property
    def current_adc(self):
        """Return the currently used ADC"""
        curradc = from_currsys(self.meta['current_adc'])
        if not curradc:
            return False
        return self.adcs[curradc]

    @property
    def display_name(self):
        return f'{self.meta["name"]} : ' \
               f'[{from_currsys(self.meta["current_adc"])}]'

    def __getattr__(self, item):
        return getattr(self.current_adc, item)

    def get_table(self):
        """Create a table of ADCs with maximimum througput"""
        names = list(self.adcs.keys())
        adcs = self.adcs.values()
        tmax = np.array([adc.data['transmission'].max() for adc in adcs])

        tbl = Table(names=["name", "max_transmission"],
                    data=[names, tmax])
        return tbl



import numpy as np


from astropy.io import fits
from astropy.table import Table

from .image_plane_utils import add_table_to_imagehdu, add_imagehdu_to_imagehdu

from ..base_classes import ImagePlaneBase
from .. import rc
from .. import utils


class ImagePlane(ImagePlaneBase):
    """
    A class to act as a canvas onto which to project `Source` images or tables

    Parameters
    ----------
    header : `fits.Header`
        Must contain a valid WCS

    .. todo: Write the code to deal with a canvas larger than max_segment_size

    Examples
    --------
    ::

        from astropy.table import Table
        from scopesim.optics import image_plane as imp

        my_point_source_table = Table(names=["x", "y", "flux"],
                                      data=[(0,  1,  2)*u.mm,
                                            (0, -5, 10)*u.mm,
                                            (100,50,25)*u.ph/u.s])

        hdr = imp.make_image_plane_header([my_point_source_table],
                                           pixel_size=0.015*u.mm)
        img_plane = imp.ImagePlane(hdr)
        img_plane.add(my_point_source_table)

        print(img_plane.image)

    """

    def __init__(self, header, **kwargs):

        max_seg_size = rc.__config__["!SIM.computing.max_segment_size"]
        self.meta = {"SIM_MAX_SEGMENT_SIZE" : max_seg_size}
        self.meta.update(kwargs)
        self.id = header["IMGPLANE"] if "IMGPLANE" in header else 0

        if not any([utils.has_needed_keywords(header, s)
                    for s in ["", "D", "S"]]):
            raise ValueError("header must have a valid image-plane WCS: {}"
                             "".format(dict(header)))

        image = np.zeros((header["NAXIS2"]+1, header["NAXIS1"]+1))
        self.hdu = fits.ImageHDU(data=image, header=header)

    def add(self, hdus_or_tables, sub_pixel=None, spline_order=None, wcs_suffix=""):
        """
        Add a projection of an image or table files to the canvas

        .. note::
            If a Table is provided, it must include the following columns:
            `x_mm`, `y_mm`, and `flux`.

            Units for the columns should be provided in the
            <Table>.unit attribute or as an entry in the table's meta dictionary
            using this syntax: <Table>.meta["<colname>_unit"] = <unit>.

            For example::

              tbl["x"].unit = u.arcsec   # or
              tbl.meta[x_unit"] = "deg"

            If no units are given, default units will be assumed. These are:

            - `x`, `y`: `arcsec`
            - `flux` : `ph / s / pix`

        Parameters
        ----------
        hdus_or_tables : `fits.ImageHDU` or `astropy.Table`
            The input to be projected onto the image plane. See above.

        sub_pixel : bool, optional
            Default is False. Dictates if point files should be projected with
            sub-pixel shifts or not. Accounting for sub-pixel shifts is approx.
            5x slower.

        spline_order : int, optional
            Order of spline interpolations used in ``scipy.ndimage`` functions
            ``zoom`` and ``rotate``.

        wcs_suffix : str, optional
            Default "". For sky coords - "" or "S", Detector coords - "D"

        """
        if sub_pixel is None:
            sub_pixel = utils.from_currsys("!SIM.sub_pixel.flag")
        if spline_order is None:
            spline_order = utils.from_currsys("!SIM.computing.spline_order")

        if isinstance(hdus_or_tables, (list, tuple)):
            for hdu_or_table in hdus_or_tables:
                self.add(hdu_or_table, sub_pixel, spline_order, wcs_suffix)
        else:
            if isinstance(hdus_or_tables, Table):
                self.hdu.header["COMMENT"] = "Adding files from table"
                self.hdu = add_table_to_imagehdu(hdus_or_tables, self.hdu,
                                                 sub_pixel, wcs_suffix)
            elif isinstance(hdus_or_tables, fits.ImageHDU):
                self.hdu.header["COMMENT"] = "Adding files from table"
                self.hdu = add_imagehdu_to_imagehdu(hdus_or_tables, self.hdu,
                                                    spline_order, wcs_suffix)

    @property
    def header(self):
        return self.hdu.header

    @property
    def data(self):
        return self.hdu.data

    @data.setter
    def data(self, new_data):
        self.hdu.data = new_data

    @property
    def image(self):
        return self.data

    def view(self, sub_pixel):
        # for consistency with FieldOfView
        return self.data



import logging

import numpy as np
from astropy import units as u
from astropy.table import Table

from ..base_classes import FOVSetupBase
from .effects import Effect
from .apertures import ApertureMask
from .. import utils
from ..optics.image_plane_utils import header_from_list_of_xy, calc_footprint

__all__ = ["DetectorList", "DetectorWindow"]


class DetectorList(Effect):
    """
    A description of detector positions and properties

    The list of detectors must have the following table columns
    ::

        id   x_cen   y_cen  x_size  y_size  pixel_size  angle    gain

    where:
    * "id" is a reference id for the chip (fits header EXTNAME)
    * "x_cen" and "y_cen" [mm] are the physical coordinates of centre of the chip on the detector plane
    * "x_size", "y_size" [mm, pixel] are the width/height of the chip
    * "pixel_size" [mm] is the physical size of pixels in the detector
    * "angle" [deg] is the rotation of the detector relative to the x-axis
    * "gain" [e-/ADU] is the conversion factor for electrons (photons) to ADUs

    The units for each column (except ``id``) must be given in the meta data
    using the format ``<colname>_unit``. E.g. ``x_size_unit``.
    See examples below.

    .. note::
       Currently only the units specified below are accepted.

       For ``x(y)_size_unit``, acceptable units are ``mm``, ``pixel``

    Parameters
    ----------
    filename : str, optional
        Filename of the ASCII file with the detector description. See examples

    array_dict : dict
        Dict containing the detector description. See examples

    image_plane_id : int
        Which image plane the detector will look at (generally 0)

    Examples
    --------
    With the ``array_dict`` feature
    ::

        -   name: single_detector
            class: DetectorList
            kwargs:
                image_plane_id : 0
                array_dict:
                    id: [1]
                    x_cen: [0.]
                    y_cen: [0.]
                    x_size: [5.12]
                    y_size: [5.12]
                    pixel_size: [0.01]
                    angle: [0.]
                    gain: [1.0]
                x_cen_unit: mm
                y_cen_unit: mm
                x_size_unit: mm
                y_size_unit: mm
                pixel_size_unit: mm
                angle_unit: deg
                gain_unit: electron/adu


    Or referring to a table contained in a seperate ASCII file
    ::

        - name : full_detector_array
          class : DetectorList
          kwargs :
            filename : "detecotr_list.dat"
            active_detectors : [1, 3]
            image_plane_id : 0

    where the file detecotr_list.dat contains the following information
    ::

        # x_cen_unit : mm
        # y_cen_unit : mm
        # x_size_unit : pix
        # y_size_unit : pix
        # pixel_size_unit : mm
        # angle_unit : deg
        # gain_unit : electron/adu
        #
        id   x_cen   y_cen  x_size  y_size  pixel_size  angle    gain
        1   -63.94    0.00    4096    4096       0.015    0.0     1.0
        2     0.00    0.00    4096    4096       0.015   90.0     1.0
        3    63.94    0.00    4096    4096       0.015  180.0     1.0

    """
    def __init__(self, **kwargs):
        super(DetectorList, self).__init__(**kwargs)
        params = {"z_order": [90, 290, 390, 490],
                  "pixel_scale": "!INST.pixel_scale",      # arcsec
                  "active_detectors": "all",
                  "report_plot_include": True,
                  "report_table_include": True}
        self.meta.update(params)
        self.meta.update(kwargs)

        # for backwards compatibility
        new_colnames = {"xhw": "x_size", "yhw": "y_size", "pixsize": "pixel_size"}
        mult_cols = {"xhw": 2., "yhw": 2., "pixsize": 1.}
        if isinstance(self.table, Table):
            for col in new_colnames:
                if col in self.table.colnames:
                    self.table[col] = self.table[col] * mult_cols[col]
                    self.table.rename_column(col, new_colnames[col])
        if not "x_size_unit" in self.meta and "xhw_unit" in self.meta:
            self.meta["x_size_unit"] = self.meta["xhw_unit"]
        if not "y_size_unit" in self.meta and "yhw_unit" in self.meta:
            self.meta["y_size_unit"] = self.meta["yhw_unit"]

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, FOVSetupBase):

            hdr = self.image_plane_header
            x_mm, y_mm = calc_footprint(hdr, "D")
            pixel_size = hdr["CDELT1D"]              # mm
            pixel_scale = (kwargs.get("pixel_scale", self.meta["pixel_scale"]))   # ["]
            pixel_scale = utils.from_currsys(pixel_scale)
            x_sky = x_mm * pixel_scale / pixel_size  # x["] = x[mm] * ["] / [mm]
            y_sky = y_mm * pixel_scale / pixel_size  # y["] = y[mm] * ["] / [mm]

            obj.shrink(axis=["x", "y"], values=([min(x_sky), max(x_sky)],
                                                [min(y_sky), max(y_sky)]))
            obj.detector_limits = {"xd_min": min(x_mm),
                                   "xd_max": max(x_mm),
                                   "yd_min": min(y_mm),
                                   "yd_max": max(y_mm)}

        return obj

    def fov_grid(self, which="edges", **kwargs):
        """Returns an ApertureMask object. kwargs are "pixel_scale" [arcsec]"""
        logging.warning("DetectorList.fov_grid will be depreciated in v1.0")
        aperture_mask = None
        if which == "edges":
            self.meta.update(kwargs)
            self.meta = utils.from_currsys(self.meta)

            hdr = self.image_plane_header
            x_mm, y_mm = calc_footprint(hdr, "D")
            pixel_size = hdr["CDELT1D"]              # mm
            pixel_scale = self.meta["pixel_scale"]   # ["]
            x_sky = x_mm * pixel_scale / pixel_size  # x["] = x[mm] * ["] / [mm]
            y_sky = y_mm * pixel_scale / pixel_size  # y["] = y[mm] * ["] / [mm]

            aperture_mask = ApertureMask(array_dict={"x": x_sky, "y": y_sky},
                                         pixel_scale=pixel_scale)

        return aperture_mask

    @property
    def image_plane_header(self):
        tbl = self.active_table
        pixel_size = np.min(utils.quantity_from_table("pixel_size", tbl, u.mm))
        x_unit = utils.unit_from_table("x_size", tbl, u.mm)
        y_unit = utils.unit_from_table("y_size", tbl, u.mm)

        xcen = tbl["x_cen"].data.astype(float)
        ycen = tbl["y_cen"].data.astype(float)
        dx = 0.5 * tbl["x_size"].data.astype(float)
        dy = 0.5 * tbl["y_size"].data.astype(float)

        scale_unit = 1        # either unitless to retain
        if "pix" in x_unit.name:
            scale_unit = u.mm / u.pix
            dx *= pixel_size.value
            dy *= pixel_size.value

        x_det_min = np.min(xcen - dx) * x_unit * scale_unit
        x_det_max = np.max(xcen + dx) * x_unit * scale_unit
        y_det_min = np.min(ycen - dy) * y_unit * scale_unit
        y_det_max = np.max(ycen + dy) * y_unit * scale_unit

        x_det = [x_det_min.to(u.mm).value, x_det_max.to(u.mm).value]
        y_det = [y_det_min.to(u.mm).value, y_det_max.to(u.mm).value]

        pixel_size = pixel_size.to(u.mm).value
        hdr = header_from_list_of_xy(x_det, y_det, pixel_size, "D")
        hdr["IMGPLANE"] = self.meta["image_plane_id"]

        return hdr

    @property
    def active_table(self):
        if self.meta["active_detectors"] == "all":
            tbl = self.table
        elif isinstance(self.meta["active_detectors"], (list, tuple)):
            mask = [det_id in self.meta["active_detectors"]
                    for det_id in self.table["id"]]
            tbl = self.table[mask]
        else:
            raise ValueError("Could not determine which detectors are active: "
                             "{}, {}, ".format(self.meta["active_detectors"],
                                               self.table))
        tbl = utils.from_currsys(tbl)

        return tbl

    def detector_headers(self, ids=None):
        if ids is not None and all([isinstance(ii, int) for ii in ids]):
            self.meta["active_detectors"] = list(ids)

        tbl = utils.from_currsys(self.active_table)
        hdrs = []
        for row in tbl:
            pixel_size = row["pixel_size"]
            xcen, ycen = float(row["x_cen"]), float(row["y_cen"])
            dx, dy = 0.5 * float(row["x_size"]), 0.5 * float(row["y_size"])

            if "pix" in self.meta["x_cen_unit"]:
                xcen, ycen = xcen * pixel_size, ycen * pixel_size
            if "pix" in self.meta["x_size_unit"]:
                dx, dy = dx * pixel_size, dy * pixel_size

            hdr = header_from_list_of_xy([xcen-dx, xcen+dx], [ycen-dy, ycen+dy],
                                         pixel_scale=pixel_size, wcs_suffix="D")
            if abs(row["angle"]) > 1E-4:
                sang = np.sin(row["angle"] / 57.29578)
                cang = np.cos(row["angle"] / 57.29578)
                hdr["PC1_1"] = cang
                hdr["PC1_2"] = sang
                hdr["PC2_1"] = -sang
                hdr["PC2_2"] = cang

            # hdr["GAIN"] = row["gain"]
            if "id" in row:
                hdr["DET_ID"] = row["id"]
                hdr["EXTNAME"] = f'DET_{row["id"]}'

            row_dict = {col: row[col] for col in row.colnames}
            hdr.update(row_dict)
            hdrs += [hdr]

        return hdrs

    def plot(self):
        import matplotlib.pyplot as plt
        plt.gcf().clf()

        for hdr in self.detector_headers():
            x_mm, y_mm = calc_footprint(hdr, "D")
            x_cen, y_cen = np.average(x_mm), np.average(y_mm)
            x_mm = list(x_mm) + [x_mm[0]]
            y_mm = list(y_mm) + [y_mm[0]]
            plt.gca().plot(x_mm, y_mm)
            plt.gca().text(x_cen, y_cen, hdr["ID"])

        plt.gca().set_aspect("equal")
        plt.ylabel("Size [mm]")

        return plt.gcf()


class DetectorWindow(DetectorList):
    """
    For when a full DetectorList if too cumbersome

    Parameters
    ----------
    pixel_size : float
        [mm pixel-1] Physical pixel size
    x, y : float
        [mm] Position of window centre relative to optical axis
    width, height=None : float
        [mm] Dimensions of window. If height is None, height=width
    angle : float, optional
        [deg] Rotation of window
    gain : float, optional
        [ADU/e-]
    units : str, optional
        [mm, pixel] Default "mm". Sets the input parameter units.
        If ``"pixel"``, (``x``, ``y``, ``width``, ``height``) are multiplied
        by ``pixel_size``

    """
    def __init__(self, pixel_size, x, y, width, height=None, angle=0, gain=1,
                 units="mm", **kwargs):

        if height is None:
            height = width

        params = {"orig_units": units,
                  "x_cen_unit": units,
                  "y_cen_unit": units,
                  "x_size_unit": units,
                  "y_size_unit": units,
                  "pixel_size_unit": "mm",
                  "angle_unit": "deg",
                  "gain_unit": "electron/adu",
                  "image_plane_id": 0}
        params.update(kwargs)

        tbl = Table(data=[[0], [x], [y], [width], [height],
                          [angle], [gain], [pixel_size]],
                    names=["id", "x_cen", "y_cen", "x_size", "y_size",
                           "angle", "gain", "pixel_size"])
        tbl.meta.update(params)

        super(DetectorWindow, self).__init__(table=tbl, **params)



"""Effects describing observing strategies

- ChopNodCombiner: simulate chop-nod cycle
"""
import numpy as np
from matplotlib import pyplot as plt

from scopesim.base_classes import DetectorBase
from scopesim.effects import Effect
from scopesim.utils import from_currsys, check_keys


class ChopNodCombiner(Effect):
    """
    Creates and combines 4 images for each of the chop/nod positions

    - AA : original position ``(dx, dy) = (0, 0)``
    - AB : chop position ``(dx, dy) = chop_offsets``
    - BA : nod position ``(dx, dy) = nod_offsets``
    - BB : chop-nod position ``(dx, dy) = nod_offsets + chop_offsets``

    Images are combined using::

        im_combined = (AA - AB) - (BA - BB)

    If no ``nod_offset`` is given, it is set to the inverse of ``chop_offset``.

    ``ChopNodCombiner`` is a detector effect and should be placed last in the
     detector yaml (after the noise effects).

    Keyword arguments
    -----------------
    chop_offsets : tuple, optional
        [arcsec] (dx, dy) offset of chop position relative to AA
    nod_offsets : tuple, optional
        [arcsec] (dx, dy) offset of nod position relative to AA

    Example yaml entry
    ------------------
    ::

        name: perpendicular_chop_nod_slanted_pattern
        description: chop throw to (+5, 0) and nod throw to (-7, +10) arcsec
        class: ChopNodCombiner
        include: True
        kwargs:
            pixel_scale : "!INST.pixel_scale"
            chop_offsets : (5, 0)
            nod_offsets : (-7, 10)

    """

    def __init__(self, **kwargs):
        check_keys(kwargs, ["chop_offsets", "pixel_scale"])

        super().__init__(**kwargs)
        params = {"chop_offsets": None,
                  "nod_offsets": None,
                  "pixel_scale": None,
                  "include": True,
                  "z_order": [863]}
        self.meta.update(params)
        self.meta.update(kwargs)

    def apply_to(self, obj, **kwargs):
        if isinstance(obj, DetectorBase):
            chop_offsets = from_currsys(self.meta["chop_offsets"])
            nod_offsets = from_currsys(self.meta["nod_offsets"])
            if nod_offsets is None:
                nod_offsets = -np.array(chop_offsets)

            # these offsets are in pixels, not in arcsec or mm
            pixel_scale = float(from_currsys(self.meta["pixel_scale"]))
            chop_offsets_pixel = np.array(chop_offsets) / pixel_scale
            nod_offsets_pixel = np.array(nod_offsets) / pixel_scale

            image = obj.hdu.data
            obj.hdu.data = chop_nod_image(image,
                                          chop_offsets_pixel.astype(int),
                                          nod_offsets_pixel.astype(int))

        return obj


def chop_nod_image(img, chop_offsets, nod_offsets=None):
    """Create four copies and combine in chop-nod pattern"""
    if nod_offsets is None:
        nod_offsets = tuple(-np.array(chop_offsets))

    im_AA = np.copy(img)
    im_AB = np.roll(im_AA, chop_offsets, (1, 0))
    im_BA = np.roll(im_AA, nod_offsets, (1, 0))
    im_BB = np.roll(im_BA, chop_offsets, (1, 0))

    im_comb = (im_AA - im_AB) - (im_BA - im_BB)

    return im_comb
